#ifndef __GLOBAL_TRIGGER_THROTTLE_H
#define __GLOBAL_TRIGGER_THROTTLE_H

#include <sys/types.h>
#include <sys/time.h>

#include <string>
#include <deque>

#include <boost/thread/mutex.hpp>

#include <NovaGlobalTrigger/GT.h>

#include <DAQMessages/ccpp_GTMessages.h>

/************************************************************************
 * Implementation of a single throttle pipeline for the Global Trigger  *
 *                                                                      *
 * This implimentation is done as a FIFO using std::deque               *
 * Note: This implementation uses boost::mutexes to make the deque      *
 * thread safe                                                          *
 *                                                                      *
 * There is an optional length parameter that changes the queue to be   *
 * a fixed depth.                                                       *
 ************************************************************************/

class throttlePipeline {

 public:

  throttlePipeline();
  // ~throttlePipeline(){};

  bool   pushFront(timeval evtTime);                    //! Push the time of the event onto the queue
  void   popFront();                                    //! Drop the first element of the queue

  double getRate(){return averageRate; };               //! Gets the average rate

  bool   throttleState();                               //! Returns if the pipeline should throttle or not
  void   setMaxRate(float rate){maxRate = rate;};       //! Set the rate threshold to throttle at
  double getMaxRate(){return maxRate;};                 //! Return the current value of the throttle threshold

  size_t size();                                        //! Current size of the queue
  bool   empty();                                       //! Returns true if the queue empty
  void   flush();                                       //! Flush the queue out
  void   reset();                                       //! Full reset of the queue

  int    setMaxQueueDepth(int value);                   //! Configure the maximum depth of the queue
  int    getMaxQueueDepth();
  double updateAverage();                               //! Update the rate average

  void setDebug(bool val){_debug = val;};               //! Turn on/off extra debug output
 private:

  double maxRate;                                      //! Data member for the throttle threshold (-1 to disable)
  double averageRate;                                  //! Current average rate
  double rateSum;                                      //! Current sum for the running average
  int   max_queue_depth;                               //! Maximum Size of the queue (set negative for unlimited)

  deque<timeval> theQueue;    // The pipeline implimented as a deque
  bool _debug;

}; // endclass throttlePipeline

// GT_NAMESPACE_END

#endif /* __GLOBAL_TRIGGER_THROTTLE_PIPELINE_H */
