#ifndef SNDATABUFFER__H
#define SNDATABUFFER__H

/**
* @file SNDataBuffer.h
* @brief DataBuffer class for SuperNova trigger system 
* @author Andrey Sheshukov
* @version 1
* @date 2015-04-17
*/

#include <stdint.h>
#include <deque>
#include <boost/thread/mutex.hpp>

#include "NovaSuperNova/SNBase.h"
#include "NovaSuperNova/SNDataBlock.h"
#include "NovaSuperNova/SNDataBufferIterator.h"

namespace supernova{
  /**
  * @brief DataBuffer is a queue of DataBlocks, which shares interface of VirtualDataBlock
  * When fill(t,d) method is called, it tries to find appropriate DataBlock, containing time t,
  * and fill it. 
  * If not found - extends DataBlock queue up to needed time, by creating many DataBlocks with 
  * duration= _blockDuration and length=_blockLength
  * Provides mutex, so a buffer can be easily shared between threads.
  */
  class DataBuffer: public VirtualDataBlock, public boost::mutex
  {
    public:
      //construct empty buffer, store given block parameters
      DataBuffer(const TimeType & blockDur=100, size_t blockLen=20);
      ~DataBuffer();

      size_t nblocks() {return _blocks.size();}
      
      virtual DataType& dataAtPos(size_t n);
      virtual bool empty() const {return _blocks.empty();}
      virtual size_t size() const {return _blocks.size();}
      virtual void fill(const TimeType& t, const DataType& d);
      size_t length(){return _dataLength;}

      void setBlocksLength  (size_t len  ){_blockLength   = len;}
      void setBlocksDuration(TimeType dur){_blockDuration = dur;}
      //get front& back blocks references
      DataBlock& frontBlock(){return _blocks.front();} 
      DataBlock& backBlock() {return _blocks.back(); } 

      //add blocks to buffer, until it will cover given time point
      void extendBuffer (const TimeType& t);
      void addBlockBack ();
      void addBlockFront();
      //add existing block to buffer
      void pushBlockBack (DataBlock block) {_blocks.push_back (block); updateTimeLimits();}
      void pushBlockFront(DataBlock block) {_blocks.push_front(block); updateTimeLimits();}
      //remove block from buffer
      void popBlockBack() {_blocks.pop_back (); updateTimeLimits();}
      void popBlockFront(){_blocks.pop_front(); updateTimeLimits();}
      
      //find block, containing given point
      DataBlock& findBlock(const TimeType& t);
      //update time range, based on blocks' limits
      void updateTimeLimits();

      //iterating functionality
      //iterators, pointing to begin and end of buffer
      BufferIterator begin(){return BufferIterator(_blocks.rbegin());}
      BufferIterator end()  {return BufferIterator(_blocks.rend())  ;}
      //iterators, pointing to begin and end of n-th block
      BufferIterator beginBlock(size_t n){return BufferIterator(_blocks.rbegin()+n);}
      BufferIterator endBlock  (size_t n){return BufferIterator(_blocks.rbegin()+n+1);}
    private:
      TimeType _blockDuration;      /**< Duration of DataBlock to be created */
      size_t   _blockLength;        /**< Length of DataBlock to be created */
      size_t   _dataLength;         /**< Total length of DataBlocks */
      std::deque<DataBlock> _blocks;  /**< Collection of DataBlocks with data */
  };

}

#endif
