/* A thread dealing with Calibration Triggers */

#include <string>

// Class Def
#include <NovaGlobalTrigger/GTCalTrig.h>

#include <messagefacility/MessageLogger/MessageLogger.h>

#include <DAQDataFormats/RawTriggerHeader.h>
#include <DAQDataFormats/RawTriggerTime.h>
#include <DAQDataFormats/RawTriggerTimingMarker.h>
#include <DAQDataFormats/RawTriggerRange.h>
#include <DAQDataFormats/RawTriggerMask.h>
#include <DAQDataFormats/RawTrigger.h>

#include <DAQDataFormats/TriggerDefines.h>

#include <NovaTimingUtilities/TimingUtilities.h>

#include <NovaGlobalTrigger/GT_SendQueues.h>

#include <DAQMessages/ccpp_GTMessages.h>
#include <NovaGlobalTrigger/GTGlobals.h>
#include <NovaGlobalTrigger/GT_Stats.h>
#include <NovaGlobalTrigger/GTLongTrigs.h>

#include <boost/shared_ptr.hpp>

using namespace std;
using namespace gov::fnal::cd::rms;
using namespace novadaq;
using namespace gt;
using namespace daqdataformats;

#define TRIGGER_APP_NAME "GTCalTrig"


void GTCalTrig::configure(){
  setname(std::string("calibration pulser"));
  return;
}

void GTCalTrig::work() {

    mf::SetModuleName("GTCalTrig");
    // the (singluar!) trigger we're working with: will get pointed to
    // the right place in memory during the config step
    GTInfo *info;

    // Get the instance of the sending queues
    GT_SendQueues& theQueues = gt::GT_SendQueues::getInstance();

    // Get config stuff from global memory
    {
      boost::mutex::scoped_lock lock(gtCfgSet_mutex);
      // Grab the trigger config from global memory
      // This has to be there, at least one trigger is mandatory
      const TriggerType trigCfg = gtCfgSet.getGTConfiguration().getTriggers();

      // If there, go get the config info
      if (trigCfg.hasCalibrationSet()) {
	// if there's a Calibration config, get it, otherwise bail
	if (!trigCfg.getCalibrationSet().hasCalibration()) return;
        CalibrationType gtCalCfg = trigCfg.getCalibrationSet().getCalibration();
	// note the peculiar hybrid structure.  We still can have only
	// one calibration pulser trigger: but am trying to lay the groundwork
	// to generalize this to multiple instances: at least a random and 
	// a periodic at the same time

        // First, bail if not enabled
        if (!gtCalCfg.getEnableFlag()) return;

        // We're still here?  Configure trigger parameters
	if (gtCalCfg.getRandom()) {
	  info = &GT_Stats::getInstance().Triggers.at(TRIG_ID_SOFT_CAL_RND);
	  info->setType(TRIG_ID_SOFT_CAL_RND);
	} else {
	  info = &GT_Stats::getInstance().Triggers.at(TRIG_ID_SOFT_CAL_PER);
	  info->setType(TRIG_ID_SOFT_CAL_PER);
	}
       	info->setEnabled(gtCalCfg.getEnableFlag());
	info->setRandom(gtCalCfg.getRandom());
	info->setRate(gtCalCfg.getRate());
	info->setLength(gtCalCfg.getLength());
	info->setPrescale(gtCalCfg.getPrescale());
	info->setName(gtCalCfg.getName());
	if (info->getEnabled()) info->initMetric();
      } else { 
        return; // bail if not configured
      }
    }  // end of mutexed config access

    // get a random number seed
    srand(time(0));

    if (GT_Stats::getInstance().Triggers.at(TRIG_ID_SOFT_CAL_RND).getEnabled())
      ::mf::LogInfo("Progress") << "Random Calibration Trigger named " 
        << GT_Stats::getInstance().Triggers.at(TRIG_ID_SOFT_CAL_RND).getName() 
	<< " enabled with length " 
	<< GT_Stats::getInstance().Triggers.at(TRIG_ID_SOFT_CAL_RND).getLength()
	<< " and prescale " 
	<< GT_Stats::getInstance().Triggers.at(TRIG_ID_SOFT_CAL_RND).getPrescale() 
        << endl;
    else
      ::mf::LogInfo("Progress") << "Periodic Calibration Trigger named " 
        << GT_Stats::getInstance().Triggers.at(TRIG_ID_SOFT_CAL_PER).getName() 
	<< " enabled with length " 
	<< GT_Stats::getInstance().Triggers.at(TRIG_ID_SOFT_CAL_PER).getLength()
	<< " and prescale " 
	<< GT_Stats::getInstance().Triggers.at(TRIG_ID_SOFT_CAL_PER).getPrescale() 
        << endl;

    // loop over calibration triggers
    while (1) {
      
      // If we're not triggering, wait
      while (!Triggering) {
        boost::mutex::scoped_lock waitlock(Wait_mutex);
        Triggering_cond.wait(waitlock);
      }

      // if we were doing multiple sorts at once, we'd have to pull out
      // the right place in the GTInfo array.  Here, simply re-using
      // what we stuffed in the config step above, "info *"

      // Apply prescale
      if (!(info->incSeenCount() % info->getPrescale())) {
	// create a trigger atom
	// Version 2 is after 8/12/11's fleshing out of TriggerDefines.h
	RawTrigger trigger((version_t) 2);

	// Start with header
	RawTriggerHeader triggerHeader;

	// Next, we construct the different segments of the trigger atom
	// one at a time and stuff them into the atom

	// Set HW trigger flag
	triggerHeader.setHWFlag(0); // this is software, not hardware

	// Set source
	triggerHeader.setSource(TRIGSRC_PULSER_SYSTEM);
	if (!info->getRandom()) 
	  triggerHeader.setSourceID(TRIGSRC_ID_PULSER_P1);
	else 
	  triggerHeader.setSourceID(TRIGSRC_ID_PULSER_P2);
	triggerHeader.setSourceSubID(PartitionToSubID(partNumber));
	// In this case, it's all software, so store partition

	// Set trigger time.  In this software trigger, the two versions
	// of the trigger generation time (in nova time and timeval) are
	// the same as the trigger timing marker set below, unless there's
	// an offset in effect
	RawTriggerTime triggerTime;
	timeval now;
	gettimeofday(&now,0);
	triggerTime.setTriggerTimeStampTv(now);
	uint64_t ts_now;
	novadaq::timeutils::convertUnixTimeToNovaTime(now,ts_now);
	triggerTime.setTriggerTimeStamp(ts_now);
	triggerTime.setClockSource(CLOCKSRC_UNIX_X86_FNAL_GT_1);

	// pack our time into the trigger atom
	trigger.setTriggerTime(triggerTime);

	// Set trigger time marker
        // This is the actual spill time used to go searching for data
	RawTriggerTimingMarker triggerTimingMarker;
	timeval trigTimeTV;
	if (offsetEnable)
	  {
	    timeutils::timeval_subtract(&trigTimeTV,&now,&deltaT);
	  } else {
	  memcpy(&trigTimeTV,&now,sizeof(struct timeval));
	}
	uint64_t trigTimeNova;
	timeutils::convertUnixTimeToNovaTime(trigTimeTV,trigTimeNova);
	triggerTimingMarker.setTriggerStart_baseClock(trigTimeNova);
	// pack our time marker into the trigger atom
	trigger.setTriggerTimingMarker(triggerTimingMarker);

	// Set trigger mask
	RawTriggerMask triggerMask;
	triggerMask.setMonteCarloFlag(0);  // Not MC
	triggerMask.setTrigMod(0);         // This is the "Trigger Set"
        // confusingly named method, same as the parent's method that
        // sets this whole thing.  But, this takes a trigger bit number,
        // (trigType-1 here) and sets all the appropriate higher bits if
        // needed
        triggerMask.setTriggerMask(info->getType()-1);
	// set prescale factor
	triggerMask.setPrescale(info->getPrescale());
	
	// pack our mask into the trigger atom
	trigger.setTriggerMask(triggerMask);

	// let LT calculate info needed to chain the triggers        
	uint32_t length = info->getLength();
	GTLongTrigs LT(length);   
	// let LT adjust trigger times, loop over it and send while it works
	bool first = true;
	while (LT.adjustTrigger(&trigger) > 0) {

	  // Set trigger number.        
	  uint64_t TriggerNumberCopy = GT_Stats::getInstance().incTrigNo();
	  // keep track that we're sending a trigger of this type
          info->incSentCount();
	  triggerHeader.setTriggerNumber(TriggerNumberCopy);
	  if (first) {
	    first = false;
	    // Master Trigger number starts off the same
	    triggerHeader.setMasterTriggerNumber(TriggerNumberCopy);
	  }
	  // pack our header into the trigger atom - rest was loaded before
	  trigger.setTriggerHeader(triggerHeader);

	  // create the message, pack in the trigger atom, and send it
	  triggermessages::TriggerMessage theMessage;
	  theMessage.payload.length(4 * trigger.sizeofdata());
	  trigger.writeData(theMessage.payload.get_buffer());
	  // People are annoyed by this.  Rather than simply turning off
	  // debug messages, we've made logLevel into a message prescale.
	  if (!(info->getSeenCount()%logLevel)) 
	    LOG_DEBUG("Progress") << "Sending " << info->getName() 
				  << " " << TriggerNumberCopy << endl;

	  // Queue the Message to the central sender queues
	  theQueues.QueueTrigger("Pulser",theMessage);

        }
      } // endif prescale

      // Figure out delay.  If not random, it's easy. 
      int delay; // number of microseconds to sleep
      if (!info->getRandom()) {
        delay = (int)(1.0e6/info->getRate());
      } else {  // if it is random, do a poissonian interval with mean calRate
        delay = (int) ((1.0e6/info->getRate())*(-log((float)rand()/(float)RAND_MAX)));
      }
      if (!(info->getSeenCount()%logLevel))
        LOG_DEBUG("Progress") << "Pausing for " << delay << " us " << endl; 
      usleep(delay);
    }
}
