#include <NovaGlobalTrigger/GTInfo.h>

#include <sstream>

#include <NovaGlobalTrigger/GT_Stats.h>
#include <NovaGlobalTrigger/GTGlobals.h>
#include <NovaTimingUtilities/TimingUtilities.h>

using namespace std;
using namespace novadaq::gt;

// Stores trigger stuff in one place so we can loop over different trigs

// Note that this stuff is not mutex locked: it should all be local to 
// the one thread it's used in anyway: counting here is for prescale purposes
// not event numbers

/**
 * Constructor initializes stuff
 *
 */
GTInfo::GTInfo() 
{
  Length = 0;
  Prescale = 0;
  Enabled = false;
  Delay = 0;
  Name = "";
  SentCount = 0;
  SeenCount = 0;
  Type = daqdataformats::TRIG_ID_NULL;
  DDT = false;
  Random = false;
  Rate = 0.0;
  Queue = NULL;
}


// Destructor not needed?
//GTInfo::~GTInfo() {
// nothing at present, we dynamically allocate no memory in here to free
//}


/**
 * Returns the current Length
 *
 */
uint32_t GTInfo::getLength()
{
  return Length;
}

/**
 * Sets the Length
 *
 */
void GTInfo::setLength(uint32_t L)
{
  Length = L;
}


/**
 * Returns the current Prescale
 *
 */
uint32_t GTInfo::getPrescale()
{
  return Prescale;
}

/**
 * Sets the Prescale
 *
 */
void GTInfo::setPrescale(uint32_t P)
{
  Prescale = P;
}


/**
 * Is this trigger enabled?
 *
 */
bool GTInfo::getEnabled()
{
  return Enabled;
}

/**
 * Sets the Enabled flag
 *
 */
void GTInfo::setEnabled(bool E)
{
  Enabled = E;
}


/**
 * Returns the current Delay
 *
 */
int64_t GTInfo::getDelay()
{
  return Delay;
}

/**
 * Sets the Delay
 *
 */
void GTInfo::setDelay(int64_t D)
{
  Delay = D;
}


/**
 * Returns the current Name
 *
 */
std::string GTInfo::getName()
{
  return Name;
}

/**
 * Sets the Name
 *
 */
void GTInfo::setName(std::string N)
{
  Name = N;
}


/**
 * Returns the current trigger count: those seen before prescales
 *
 */
uint64_t GTInfo::getSeenCount()
{
  return SeenCount;
}

/**
 * returns ++SeenCount, and increments SeenCount
 *
 */
uint64_t GTInfo::incSeenCount()
{
  ++SeenCount;
  return SeenCount;
}

/**
 * clears SeenCount
 *
 */
void GTInfo::resetSeenCount()
{
  SeenCount=0;
}


/**
 * Returns the current trigger count: those sent after prescales
 *
 */
uint64_t GTInfo::getSentCount()
{
  return SentCount;
}

/**
 * returns ++SentCount, and increments SentCount
 *
 */
uint64_t GTInfo::incSentCount()
{
  GT_Stats::getInstance()._daqMonClient.setValue(TriggerRate,double(1.));
  ++SentCount;
  return SentCount;
}

/**
 * clears SentCount
 *
 */
void GTInfo::resetSentCount()
{
  SentCount=0;
}

/**
 * Returns the current lost trigger count: those lost due to throttling
 *
 */
uint64_t GTInfo::getMissedCount()
{
  return MissedCount;
}

/**
 * returns ++MissedCount, and increments MissedCount
 *
 */
uint64_t GTInfo::incMissedCount()
{
  ++MissedCount;
  return MissedCount;
}

/**
 * clears MissedCount
 *
 */
void GTInfo::resetMissedCount()
{
  MissedCount=0;
}


/**
 * Returns the current Type
 *
 */
daqdataformats::trigID GTInfo::getType()
{
  return Type;
}

/**
 * Sets the Type
 *
 */
void GTInfo::setType(daqdataformats::trigID M)
{
  Type = M;
}


/**
 * Returns the current random flag
 *
 */
bool GTInfo::getRandom()
{
  return Random;
}

/**
 * Sets the random flag
 *
 */
void GTInfo::setRandom(bool R)
{
  Random = R;
}


/**
 * Returns the current Rate
 *
 */
float GTInfo::getRate()
{
  return Rate;
}

/**
 * Sets the Rate
 *
 */
void GTInfo::setRate(float R)
{
  Rate = R;
}


/**
 * Is the trigger data drive?
 *
 */
bool GTInfo::getDDT()
{
  return DDT;
}

/**
 * Sets the DDT-ness of the trigger
 *
 */
void GTInfo::setDDT(bool D)
{
  DDT = D;
}


/**
 * Creates a ganglia rate metric for this trigger.  You'll want to set
 * Type first.
 *
 */
void GTInfo::initMetric()
{
  assert (Type!=0);
  {
    std::stringstream name;
    name << daqdataformats::getTriggerName(Type) << " rate";
    std::stringstream description;
    description << daqdataformats::getTriggerName(Type) << " rate";
    TriggerRate = GT_Stats::getInstance()._daqMonClient.createMetric<double>
      (name.str(),"Hz", nova::novadaqmonitorclient::RATE, "", 
      description.str());
  }
  if (DDT) {
    {
      std::stringstream name;
      name << daqdataformats::getTriggerName(Type) << " avg length";
      std::stringstream description;
      description << daqdataformats::getTriggerName(Type) 
                  << "average trigger length";
      DDTLength = GT_Stats::getInstance()._daqMonClient.createMetric<uint32_t>
        (name.str(),"50ns ticks", nova::novadaqmonitorclient::AVERAGE,
        "",description.str());
    }
    {
      std::stringstream name;
      name << daqdataformats::getTriggerName(Type) << " avg decision";
      std::stringstream description;
      description << daqdataformats::getTriggerName(Type) 
                  << "average trigger delay: time to decision in DDT";
      DDTDelay = GT_Stats::getInstance()._daqMonClient.createMetric<double>
        (name.str(),"ms", nova::novadaqmonitorclient::AVERAGE, "",
	description.str());
    }
    {
      std::stringstream name;
      name << daqdataformats::getTriggerName(Type) << " avg delay";
      std::stringstream description;
      description << daqdataformats::getTriggerName(Type) 
                  << "average trigger delay: total";
      DDTDecision = GT_Stats::getInstance()._daqMonClient.createMetric<double>
        (name.str(),"ms", nova::novadaqmonitorclient::AVERAGE, "",
	description.str());
    }
  }

}


/**
 * send DDT trigger length and delay to ganglia
 *
 */
void GTInfo::logDDT(uint32_t length, uint32_t decision, uint32_t delay)
{
  double msDecision = 1000.*(double)decision/(double)timeutils::NOVA_TIME_FACTOR;
  double msDelay = 1000.*(double)delay/(double)timeutils::NOVA_TIME_FACTOR;

  GT_Stats::getInstance()._daqMonClient.setValue(DDTLength,length);
  GT_Stats::getInstance()._daqMonClient.setValue(DDTDecision,msDecision);
  GT_Stats::getInstance()._daqMonClient.setValue(DDTDelay,msDelay);
}


// private methods.
