/* Deal with the program initialization */

#include <fstream>
#include <string.h>

#include <messagefacility/MessageLogger/MessageLogger.h>

#include <boost/program_options.hpp>
namespace po = boost::program_options;

#include <NovaGlobalTrigger/GTGlobals.h>
#include <NovaGlobalTrigger/GT_SendQueues.h>

#include <NovaTimingUtilities/TimingUtilities.h>
#include <NovaDAQUtilities/LocationUtils.h>

using namespace std;
using namespace novadaq;
using namespace gt;

void GTInit(int ac, char* av[]) {

//  mf::SetModuleName("GTInit");
    mf::SetContext("GTInit");

  // Parse command line options using Boost, since we already have 
  // boost linked in

  std::string config_file;
  // Declare the supported options.  This accesses partNumber without a 
  // mutex, but we're not threaded yet so it's OK.
  po::options_description desc("Allowed options");
  desc.add_options()
    ("help,h", "usage descriptions")
    ("config,c", po::value<string>(&config_file)->default_value("srt://NovaGlobalTrigger/config/GTConfig.xml"),
     "The configuration xml file to use")
    ("partition,p", po::value<int>(&partNumber)->default_value(-1), "DAQ partition to talk to")
//    ("zpartition,z", po::value<int>(&partNumber)->default_value(-1), "DAQ partition to talk to, new form")
    ;

  po::variables_map vm;
  po::store(po::parse_command_line(ac, av, desc), vm);
  po::notify(vm);    

  if (vm.count("help")) {
    cout << desc << "\n";
    exit(1);
  }

  // Get GT default config file here
  try {
    std::string fname = LocationUtils::translateURL(config_file);

    //Pulls in the XML config file

    std::string xmlString = "";
    {
      std::ifstream inStream(fname.c_str());
      if (! inStream.is_open()) {
 ::mf::LogError("Configuration | FileNotFound") 
	  << "Unable to open " << fname << "." << std::endl;
      }
      else {
	std::string line;
	while (std::getline(inStream, line)) {
	  xmlString.append(line);
	}
	inStream.close();
      }
    }

    //connects objects in XSD file to settings in XML file
    gtCfgSet.deserialize(xmlString);

    // We mutex this stuff just in case, although only one thread is 
    // active at the moment

    // Make the object of the GTConfigurationSet i.e. all the stuff in xml
    {
      boost::mutex::scoped_lock lock(gtCfgSet_mutex);
      const gt::GTConfigurationType& gtCfg = gtCfgSet.getGTConfiguration();

      // Get global config settings broken out into global variables
      if (partNumber<0) {  // if it wasn't already set
        if (gtCfg.hasPartition()) { 
          gt::PartitionType gtPartSet = gtCfg.getPartition();
          boost::mutex::scoped_lock lock(partNumber_mutex);
          partNumber = gtPartSet.getPart();
	} else {
	  partNumber = 0; // NOTE - this is really set here, not in xsd
	}
      }
      if (gtCfg.hasMilliSliceSize()) {
        gt::SliceType gtSliceSet = gtCfg.getMilliSliceSize();
        boost::mutex::scoped_lock lock(MilliSliceSize_mutex);
        MilliSliceSize = gtSliceSet.getMsSize();
      } else {
        MilliSliceSize = 10000;  // NOTE - this is really set here, not in xsd
      }
      if (gtCfg.hasOffset()) {
        gt::OffsetType gtOffsetSet = gtCfg.getOffset();
	struct tm timeinfo;
        timeinfo.tm_year = gtOffsetSet.getTv_year() - 1900;
        timeinfo.tm_mon  = gtOffsetSet.getTv_mo() - 1;
        timeinfo.tm_mday = gtOffsetSet.getTv_day();
        timeinfo.tm_hour = gtOffsetSet.getTv_hour();
        timeinfo.tm_min = gtOffsetSet.getTv_min();
        timeinfo.tm_sec = gtOffsetSet.getTv_sec();
        boost::mutex::scoped_lock lock(Offset_mutex);
        offsetEnable = gtOffsetSet.getOffsetEnable();
	if (offsetEnable) {
	  // our config time is in GMT, but mktime assumes localtime,
	  // so we have to convert back
          time_t localTimeSec = mktime(&timeinfo);

	  // Note ctime, asctime, localtime, gmtime are not threadsafe,
	  // gotta use ctime_r, asctime_r, localtime_r, gmtime_r
          struct tm gmtTimeStruct;
	  gmtime_r(&localTimeSec, &gmtTimeStruct);
          time_t gmtTimeSec = mktime(&gmtTimeStruct);
	  // calculate time zone difference, offset between localtime 
	  // and GMT.  Store it in a global so other time calls
	  // can refer to it
	  { 
	    boost::mutex::scoped_lock lock(timeZoneDiff_mutex);
            timeZoneDiff = gmtTimeSec - localTimeSec;
          }
          offsetT0.tv_sec = localTimeSec - timeZoneDiff;
          offsetT0.tv_usec = gtOffsetSet.getTv_usec();
	  struct timeval now;
	  gettimeofday(&now,0);
	  // ok - we want the program to pretend that right now is
	  // really this offsetT0 time.  So, we need to find out how
	  // long ago offsetT0 was compared to now, and subtract this
	  // delta-t from triggers to put the apparent time back into
	  // the past
	  if (timeutils::timeval_subtract(&deltaT,&now,&offsetT0) == 1)
	  {
	    ::mf::LogError("Configuration") << "Configured T0 in the future!" << endl; 
	    exit(-1);
          }
	  char ctimestr[27];
	  ctime_r(&offsetT0.tv_sec,ctimestr);
	  ::mf::LogInfo("Configuration") << "Using a T0 of: " 
	                           << ctimestr
				   << " and "
	                           << offsetT0.tv_usec << " usec," 
				   << " or " << deltaT.tv_sec << " + "
				   << ((float)deltaT.tv_usec / 1.0e6)
				   << " seconds in the past" << endl;
        }
      } else {
        offsetEnable = false;  // NOTE - this is really set here, not in xsd
      }
      if (!offsetEnable) {
	memset(&offsetT0,0,sizeof(struct timeval));
	memset(&deltaT,0,sizeof(struct timeval));
      }
      if (gtCfg.hasLog()) {
        gt::LogType gtLogLevel = gtCfg.getLog();
        boost::mutex::scoped_lock lock(logLevel_mutex);
        logLevel = gtLogLevel.getLogLevel();
	if (logLevel != 1)
	  LOG_DEBUG("Config") << "Note that you have set loglevel to " 
			     << logLevel << " so will not see every "
			     << "LogDebug trigger message" << endl;
      } else {
        logLevel = 1; // NOTE - this is really set here, not in xsd
      }

      {
	boost::mutex::scoped_lock lock(Queue_mutex);
	GT_SendQueues& theQueues = gt::GT_SendQueues::getInstance();
	int depth = -1; // Set the queue depth (-1 is unlimited)
	if (gtCfg.hasQueue()) {
	  gt::QueueType gtQueueSet = gtCfg.getQueue();
	  depth = gtQueueSet.getDepth();
	  theQueues.setPollValue(gtQueueSet.getIssuewait());
	  theQueues.setPollTimeout(gtQueueSet.getTimeout());
	  theQueues.setDebug(gtQueueSet.getDebug());
	}  else {
	  // apparently the defaults in the xml work only if there's a line
	  // in there setting something
	  theQueues.setPollValue(100);    // Set the polling time   (in microseconds)
	  theQueues.setPollTimeout(1000); // Set the time out value (in microseconds)
	  theQueues.setDebug(false);      // Set the debugging level
	}
	::mf::LogInfo("Progress") << "Trigger Queues have depth "
				  << depth
				  << " (can't change that!) polling wait "
				  << theQueues.getPollValue()
				  << " timeout parameter "
				  << theQueues.getPollTimeout()
				  << " with debug set " << theQueues.getDebug()
				  << endl;
	// I suppose the queue declarations should be in xml too.  But that's 
	// a PITA and we shouldn't be messing with them at that level anyway
	// So, hardcoded.  Also note that the MaxDepth is a per-queue thing,
	// but we're treating all the same.  The GT_SendQueues object does not
	// provide an accessor for indivdual queues.  I don't want to write one
	// simply so I can set each one to "-1" here, which is the default
	// anyway.  If for some reason someone wants finite queue depths, 
	// here's where you could start working on it.
	theQueues.AddQueue("Beam",   1);     // Beam triggers (highest priority)
	theQueues.AddQueue("Pulser", 5);     // Pulser triggers (medium priority)
	theQueues.AddQueue("DDT",    10);    // DDT triggers (low priority)
	theQueues.AddQueue("SNEWS",  15);    // SNEWS triggers (lower priority)
	theQueues.AddQueue("Other",  20);    // Other triggers (even lower priority)
	//	theQueues.ListQueues();
      }

      if (gtCfg.hasThrottle()) {
	boost::mutex::scoped_lock lock(Throttle_mutex);
        gt::ThrottleType gtThrottleSet = gtCfg.getThrottle();
        GlobalTrigThrottle.setMaxQueueDepth(gtThrottleSet.getDepth());
        GlobalTrigThrottle.setMaxRate(gtThrottleSet.getMaxrate());
	saneStats = gtThrottleSet.getSanestats();
      }  else {
	// apparently the defaults in the xml work only if there's a line
	// in there setting something
	boost::mutex::scoped_lock lock(Throttle_mutex);
        GlobalTrigThrottle.setMaxQueueDepth(100);
        GlobalTrigThrottle.setMaxRate(200.0);
	saneStats = true;
      }
      ::mf::LogInfo("Progress") << "Global Throttle has depth "
				<< GlobalTrigThrottle.getMaxQueueDepth()
				<< " and max rate "
				<< GlobalTrigThrottle.getMaxRate()
				<< " with sane stats set " << saneStats
				<< " BUT IT IS NOT YET IMPLEMENTED" << endl;

      // configure overall DDT throttle
      if (gtCfgSet.getGTConfiguration().hasDDTThrottle()) {
	boost::mutex::scoped_lock lock(Throttle_mutex);
	gt::DDTThrottleType gtDDTThrottleSet = gtCfgSet.getGTConfiguration().getDDTThrottle();
	DDTTrigThrottle.setMaxQueueDepth(gtDDTThrottleSet.getDepth());
	DDTTrigThrottle.setMaxRate(gtDDTThrottleSet.getMaxrate());
      } else {
	boost::mutex::scoped_lock lock(Throttle_mutex);
	DDTTrigThrottle.setMaxQueueDepth(100);
	DDTTrigThrottle.setMaxRate(200.0);
      }
      ::mf::LogInfo("Progress") << "Overall DDT Throttle has depth "
				<< DDTTrigThrottle.getMaxQueueDepth()
				<< " and max rate "
				<< DDTTrigThrottle.getMaxRate() << endl;

    }  // release gtCfgSet_mutex
    // LOG_DEBUG("Debug") << "XML STRING::" << xmlString;
    // Set up other globals, that aren't in the config
    { // We don't start triggering
      boost::mutex::scoped_lock lock(Triggering_mutex);
      Triggering = false;
      // notify anyone waiting on this that the state has changed
      Triggering_cond.notify_all();
    }
  }
  catch (::xsd::cxx::tree::parsing<char>& excpt) {
    ::mf::LogWarning("Configuration") << "Exception " << excpt << std::endl;
  }

  ::mf::LogInfo("Progress") << "Initialized, partition = " << partNumber << std::endl;
  return;
}
