/* A thread dealing with Manual Triggers */

/* What's a "Manual" trigger?  Two perhaps mutually exclusive things,
   for now they're together in this thread.  

   First: given some set time in the config file, issue a trigger.  Most
   useful for the simulation and testing side of things.

   Second: given a signal from run control, fire off a single-instance
   trigger, the "Big Red Button" sort of trigger.

   This dichotomy is governed by the "remote" config parameter.  If
   false, then it's the configured time.  If true, it's expecting a Big
   Red Button.  Example input to the Big Red Button sort is in the
   ManSender.cc test program.  

   */

#include <string>

#include <rms/provider/DDSConnection.h>
#include <rms/base/RmsDestination.h>
#include <rms/RmsReceiver.h>

#include <messagefacility/MessageLogger/MessageLogger.h>

#include <DAQDataFormats/RawTriggerHeader.h>
#include <DAQDataFormats/RawTriggerTime.h>
#include <DAQDataFormats/RawTriggerTimingMarker.h>
#include <DAQDataFormats/RawTriggerRange.h>
#include <DAQDataFormats/RawTriggerMask.h>
#include <DAQDataFormats/RawTrigger.h>
#include <DAQDataFormats/TriggerDefines.h>

// Class Def
#include <NovaGlobalTrigger/GTManualTrig.h>

#include <NovaTimingUtilities/TimingUtilities.h>

#include <DAQMessages/ccpp_GTMessages.h>
#include <NovaGlobalTrigger/GTGlobals.h>
#include <NovaGlobalTrigger/GT_Stats.h>
#include <NovaGlobalTrigger/GTLongTrigs.h>
#include <NovaGlobalTrigger/GT_SendQueues.h>

#include <DAQMessages/ccpp_DDTMessages.h>

#include <boost/shared_ptr.hpp>

using namespace std;
using namespace gov::fnal::cd::rms;
using namespace novadaq;
using namespace gt;

#define TRIGGER_APP_NAME "GTManualTrig"

void GTManualTrig::configure(){
  setname(std::string("manual trigger"));
  return;
}

void GTManualTrig::work() {

    mf::SetModuleName("GTManualTrig");

    // Get the instance of the sending queues
    GT_SendQueues& theQueues = gt::GT_SendQueues::getInstance();

    bool remote = false;
    // Get config stuff from global memory
    struct timeval manualTrigTime;
    { 
      boost::mutex::scoped_lock lock(gtCfgSet_mutex);
      // Grab the trigger config from global memory
      // This has to be there, at least one trigger is mandatory
      const TriggerType trigCfg = gtCfgSet.getGTConfiguration().getTriggers();

      // If there, go get the config info
      if (trigCfg.hasManualSet()) {
	// if there's Manual config stuff, get it
        if (!trigCfg.getManualSet().hasManual()) return;
        ManualType gtManualCfg = trigCfg.getManualSet().getManual();

        // First, bail if not enabled
        if (!gtManualCfg.getEnableFlag()) return;

        // We're still here?  Configure trigger parameters
	GTInfo *info = &GT_Stats::getInstance().Triggers.at(daqdataformats::TRIG_ID_SOFT_MANUAL);
        info->setLength(gtManualCfg.getLength());
        info->setPrescale(gtManualCfg.getPrescale());
	info->setName(gtManualCfg.getName());
	info->setEnabled(true);
	info->setType(daqdataformats::TRIG_ID_SOFT_MANUAL);
	info->initMetric();
	info->Queue = new throttlePipeline;
	info->Queue->setMaxQueueDepth(gtManualCfg.getDepth());
	info->Queue->setMaxRate(gtManualCfg.getMaxrate());
        struct tm timeinfo;   // Remember, we work in GMT!  But this is localtime
        timeinfo.tm_year = gtManualCfg.getTv_year() - 1900;
        timeinfo.tm_mon  = gtManualCfg.getTv_mo() - 1;
        timeinfo.tm_mday = gtManualCfg.getTv_day();
        timeinfo.tm_hour = gtManualCfg.getTv_hour();
        timeinfo.tm_min = gtManualCfg.getTv_min();
        timeinfo.tm_sec = gtManualCfg.getTv_sec();
        manualTrigTime.tv_sec = mktime(&timeinfo) - timeZoneDiff;
        manualTrigTime.tv_usec = gtManualCfg.getTv_usec();

	remote = gtManualCfg.getRemote();
      } else {
        return; // bail if not configured
      }
    }  // end of mutexed config access

    if (remote) 
      ::mf::LogInfo("Progress") << "Manual Trigger named " 
				<< GT_Stats::getInstance().Triggers.at(daqdataformats::TRIG_ID_SOFT_MANUAL).getName()
				<< " enabled in remote mode" 
				<< " with prescale "
				<< GT_Stats::getInstance().Triggers.at(daqdataformats::TRIG_ID_SOFT_MANUAL).getPrescale()
				<< " queue depth " 
				<< GT_Stats::getInstance().Triggers.at(daqdataformats::TRIG_ID_SOFT_MANUAL).Queue->getMaxQueueDepth()
				<< " max rate "
				<< GT_Stats::getInstance().Triggers.at(daqdataformats::TRIG_ID_SOFT_MANUAL).Queue->getMaxRate()
				<< endl;
    else
      ::mf::LogInfo("Progress") << "Manual Trigger named " 
				<< GT_Stats::getInstance().Triggers.at(daqdataformats::TRIG_ID_SOFT_MANUAL).getName()
				<< " enabled in one shot mode" << endl;

    // create the RMS connection for the remote manual trigger
    // Must use the same partition number!
    boost::shared_ptr<provider::DDSConnection>
        rmsManConnection(new provider::DDSConnection("GTManListener",
                                    partNumber));
    // create the destination that represents where we will receive
    // messages from.  Re-using the DDT message type. 

    base::RmsDestination receiveDest("ManServerMailbox",
                                     base::RmsDestination::TRIGGER_CHANNEL);
    // create a receiver with the connection object and destination
    RmsReceiver<provider::DDSConnection, ddtmessages::DDTTriggerMessage>
        requestReceiver(rmsManConnection, receiveDest);


    // Here we would loop to take Big Red Button triggers: but only once
    // if the need is to issue the xml-configured time and exit

    bool loop = true;
    while (loop) {
      // If we're not triggering, wait
      while (!Triggering) {
        boost::mutex::scoped_lock waitlock(Wait_mutex);
        Triggering_cond.wait(waitlock);
      }

      ddtmessages::DDTTriggerMessage ddtMessage;
      if (!remote) 
	loop = false;    // let's just do this once
      else //grab msgs and store in ddtMessage
	requestReceiver.receiveMessage(ddtMessage);	

      GTInfo *info = &GT_Stats::getInstance().Triggers.at(daqdataformats::TRIG_ID_SOFT_MANUAL);

      // Apply prescale
      if (!(info->incSeenCount() % info->getPrescale())) {

	// Determine when this trigger came in (to the GT)
        timeval trigArrival;
        gettimeofday(&trigArrival,0);

        // Calculation of rolling trigger rates
        // Push this time value into the global throttle
        //GlobalTrigThrottle.pushFront(trigArrival);

        // Push this time value into the trigger specific throttle
        info->Queue->pushFront(trigArrival);
	
	// Check to see if our average trigger rate has spiked above our maximum rate
	if(info->Queue->throttleState()){
  	  // We have execeeded our throttle on this trigger
	  // TODO: add global rate check too

	  // Issue a warning message
	  ::mf::LogWarning("Trigger Rate") << "Specific Trigger Rate Average: " << info->Queue->getRate() 
					   << "\nSpecific Trigger rate exceeds maximum, throttling trigger output.\n" 
					   << "Type of last trigger: " 
					   << daqdataformats::getTriggerShortName(daqdataformats::TRIG_ID_SOFT_MANUAL)
					   << std::endl;
	
	  // We back off and cowardly refuse to issue the trigger
	  // We pop the lastest trigger (the one we are not going to issue) off the throttle queue
	  // which has the effect of pushing the rolling average down the next time through 
	  info->Queue->popFront();

	  // We bounce out since we aren't going to issue this trigger

	  info->incMissedCount();

	  continue;
	}


	// create a trigger atom
	// Version 2 is after 8/12/11's fleshing out of TriggerDefines.h
	daqdataformats::RawTrigger trigger((daqdataformats::version_t) 2);

	// Start with header
	daqdataformats::RawTriggerHeader triggerHeader;

	// Next, we construct the different segments of the trigger atom
	// one at a time and stuff them into the atom

	// Set HW trigger flag
	triggerHeader.setHWFlag(0); // this is software, not hardware.

	// Set source
	triggerHeader.setSource(daqdataformats::TRIGSRC_MANUAL);
	triggerHeader.setSourceID(daqdataformats::TRIGSRC_ID_GT_NDOS);
	triggerHeader.setSourceSubID(daqdataformats::PartitionToSubID(partNumber));
	// In this case, it's all software, so store partition
    
	// Set trigger time.
	daqdataformats::RawTriggerTime triggerTime;

	// This is the time the trigger was generated
	timeval now;
	gettimeofday(&now,0);
	triggerTime.setTriggerTimeStampTv(now);
	uint64_t ts_now;
	novadaq::timeutils::convertUnixTimeToNovaTime(now,ts_now);
	triggerTime.setTriggerTimeStamp(ts_now);
	triggerTime.setClockSource(daqdataformats::CLOCKSRC_UNIX_X86_FNAL_GT_1);

	// pack our time into the trigger atom
	trigger.setTriggerTime(triggerTime);

	// Set trigger time marker
	// This is the actual spill time used to go searching for data
	daqdataformats::RawTriggerTimingMarker triggerTimingMarker;
	uint64_t trigTimeNova;
	if (remote)
	  trigTimeNova = ddtMessage.start_time;
	else {
	  timeval trigTimeTV;
	  if (offsetEnable)
	    {
	      timeutils::timeval_subtract(&trigTimeTV,&manualTrigTime,&deltaT);
	    } else {
	    memcpy(&trigTimeTV,&manualTrigTime,sizeof(struct timeval));
	  }
	  timeutils::convertUnixTimeToNovaTime(trigTimeTV,trigTimeNova);
	}
	triggerTimingMarker.setTriggerStart_baseClock(trigTimeNova);
	// pack our time marker into the trigger atom
	trigger.setTriggerTimingMarker(triggerTimingMarker);

	// Set trigger mask
	daqdataformats::RawTriggerMask triggerMask;
	triggerMask.setMonteCarloFlag(0);  // Not MC
	triggerMask.setTrigMod(0);         // This is the "Trigger Set"
	triggerMask.setTriggerMask(info->getType()-1);
	// set prescale factor
	// Add in those we've missed since last trigger due to throttling.
	// The algebra is to account for the fact that if we throttled 
	// something that was already prescaled, we didn't just lose one 
	// more, but the last n as well
	triggerMask.setPrescale(info->getPrescale() * 
				(1 + info->getMissedCount()));
	info->resetMissedCount();
	
	// pack our mask into the trigger atom
	trigger.setTriggerMask(triggerMask);

	// let LT calculate info needed to chain the triggers        
	uint32_t length;
	if (remote)
	  length = ddtMessage.duration;  // use length passed in
	else
	  length = info->getLength();
	GTLongTrigs LT(length);  // use length in config
	// let LT adjust trigger times, loop over it and send while it
	// works
	bool first = true;
	while (LT.adjustTrigger(&trigger) > 0) {

	  // Set trigger number.        
	  uint64_t TriggerNumberCopy = GT_Stats::getInstance().incTrigNo();
	  // keep track that we're sending a trigger of this type
	  info->incSentCount();
	  triggerHeader.setTriggerNumber(TriggerNumberCopy);
	  if (first) {
	    first = false;
	    // Master Trigger number starts off the same
	    triggerHeader.setMasterTriggerNumber(TriggerNumberCopy);
	  }
	  // pack our header into the trigger atom - rest was loaded before
	  trigger.setTriggerHeader(triggerHeader);

	  // create the message, pack in the trigger atom, and send it
	  triggermessages::TriggerMessage theMessage;
	  theMessage.payload.length(4 * trigger.sizeofdata());
	  trigger.writeData(theMessage.payload.get_buffer());
	  // People are annoyed by this.  Rather than simply turning off
	  // debug messages, we've made logLevel into a message prescale.
	  if (!(info->getSeenCount()%logLevel)) 
	    LOG_DEBUG("Progress") << "Sending Manual trigger number "
				  << TriggerNumberCopy << endl;

	  // Queue the Message to the central sender queues
	  theQueues.QueueTrigger("Other",theMessage);

        }
      } // endif prescale

    }  // end of loop

}
