/* A thread dealing with SNEWS Triggers */
/* Triggers come from a SpillServer-style message socket */

#include <string>

#include <rms/provider/DDSConnection.h>
#include <rms/base/RmsDestination.h>
#include <rms/RmsReceiver.h>

#include <messagefacility/MessageLogger/MessageLogger.h>

#include <DAQDataFormats/RawTriggerHeader.h>
#include <DAQDataFormats/RawTriggerTime.h>
#include <DAQDataFormats/RawTriggerTimingMarker.h>
#include <DAQDataFormats/RawTriggerRange.h>
#include <DAQDataFormats/RawTriggerMask.h>
#include <DAQDataFormats/RawTrigger.h>

#include <DAQDataFormats/TriggerDefines.h>

#include <NovaTimingUtilities/TimingUtilities.h>

#include <NovaSNEWSInterface/SNEWSMessageInfo.h>
#include <DAQMessages/ccpp_SNEWSMessages.h>

#include <DAQMessages/ccpp_GTMessages.h>
#include <NovaGlobalTrigger/GTGlobals.h>
#include <NovaGlobalTrigger/GT_Stats.h>
#include <NovaGlobalTrigger/GTLongTrigs.h>
#include <NovaGlobalTrigger/GTSNEWSTrig.h>

#include <NovaGlobalTrigger/GT_SendQueues.h>

#include <boost/shared_ptr.hpp>

using namespace std;
using namespace gov::fnal::cd::rms;
using namespace novadaq;
using namespace gt;
using namespace daqdataformats;

#define TRIGGER_APP_NAME "GTSNEWSTrig"

void GTSNEWSTrig::configure() { 
    mf::SetModuleName("GTSNEWSTrig");
    setname(std::string("data driven triggers"));

    {
      boost::mutex::scoped_lock lock(gtCfgSet_mutex);

      bool somethingEnabled = false;

      // Grab the trigger config from global memory
      // This has to be there, at least one trigger is mandatory
      const TriggerType trigCfg = gtCfgSet.getGTConfiguration().getTriggers();

      // Get SNEWS config stuff from global memory if it is there
      if (trigCfg.hasSNEWSSet()) {
	if (trigCfg.getSNEWSSet().hasSNEWS()) {
	  int snewsCfgCount = trigCfg.getSNEWSSet().getSNEWSCount();
	  LOG_INFO("Progress") << "There are " << snewsCfgCount << " SNEWS Triggers listed" << endl;
	  for(int indx = 0 ; indx < snewsCfgCount; ++indx) {
            SNEWSType gtSNEWSCfg = trigCfg.getSNEWSSet().getSNEWS(indx);
	    trigID trigType = TRIG_ID_NULL;
	    if (gtSNEWSCfg.getName() == "SNEWSTrig")
	      trigType = TRIG_ID_SNEWS_TRIGGER;
	    else if (gtSNEWSCfg.getName() == "SNEWSBeatSlow") 
	      trigType = TRIG_ID_SNEWS_BEAT_SLOW;
	    else if (gtSNEWSCfg.getName() == "SNEWSBeatFast")
	      trigType = TRIG_ID_SNEWS_BEAT_FAST;

	    // copy stuff to the GTInfo element for this trigger: bit #0
	    GTInfo *info = &GT_Stats::getInstance().Triggers.at(trigType);
	    info->setEnabled(gtSNEWSCfg.getEnableFlag());
	    info->setLength(gtSNEWSCfg.getLength());
	    info->setPrescale(gtSNEWSCfg.getPrescale());
	    info->setDelay(gtSNEWSCfg.getDelay());
	    info->setName(gtSNEWSCfg.getName());
	    info->setType(trigType);
	    if (info->getEnabled()) {
	      info->initMetric();
	      somethingEnabled = true;
	    }
	  }// end of loop over SNEWS triggers
	}// end of if it has SNEWS objects
      }// end of it it has SNEWS Set object

      // Bail if no SNEWS signals are enabled
      if (!somethingEnabled) return;

      // create overall SNEWS metrics only if we reach here, needs actual function
      GT_Stats::getInstance().initSNEWSMetrics();
    }  // end of mutexed config access

    // Log which triggers are enabled
    if (GT_Stats::getInstance().Triggers.at(TRIG_ID_SNEWS_TRIGGER).getEnabled()) 
      ::mf::LogInfo("Progress") << "SNEWS Trigger named "
	<< GT_Stats::getInstance().Triggers.at(TRIG_ID_SNEWS_TRIGGER).getName() 
	<< " enabled with length " 
	<< GT_Stats::getInstance().Triggers.at(TRIG_ID_SNEWS_TRIGGER).getLength()
        << ", delay " 
	<< GT_Stats::getInstance().Triggers.at(TRIG_ID_SNEWS_TRIGGER).getDelay()
        << ", and prescale " 
	<< GT_Stats::getInstance().Triggers.at(TRIG_ID_SNEWS_TRIGGER).getPrescale() 
	<< endl;

    if (GT_Stats::getInstance().Triggers.at(TRIG_ID_SNEWS_BEAT_FAST).getEnabled()) 
      ::mf::LogInfo("Progress") << "SNEWS Trigger named "
	<< GT_Stats::getInstance().Triggers.at(TRIG_ID_SNEWS_BEAT_FAST).getName() 
	<< " enabled with length " 
	<< GT_Stats::getInstance().Triggers.at(TRIG_ID_SNEWS_BEAT_FAST).getLength()
        << ", delay " 
	<< GT_Stats::getInstance().Triggers.at(TRIG_ID_SNEWS_BEAT_FAST).getDelay()
        << ", and prescale " 
	<< GT_Stats::getInstance().Triggers.at(TRIG_ID_SNEWS_BEAT_FAST).getPrescale() 
	<< endl;
}// end configure

void GTSNEWSTrig::work() {
    // Get the instance of the sending queues
    GT_SendQueues& theQueues = gt::GT_SendQueues::getInstance();

    //TODO Check the RMS stuff for SNEWS, is this what Andrew said he would put in a common function?

    // create the RMS connection for SNEWS, on the null partition
    boost::shared_ptr<provider::DDSConnection>
        rmsSpillConnection(new provider::DDSConnection("GTSNEWSListener",
                                    base::RmsDestination::NULL_PARTITION));

    // create the destination that represents where we will receive
    // messages from
    base::RmsDestination receiveDest("SNEWSServerMailbox",
                                     base::RmsDestination::TRIGGER_CHANNEL);

    // create a receiver with the connection object and destination
    RmsReceiver<provider::DDSConnection,
        nsimessages::SNEWSMessage>
        requestReceiver(rmsSpillConnection, receiveDest);

    while (true) {
      // If we're not triggering, wait
      while (!Triggering) {
        boost::mutex::scoped_lock waitlock(Wait_mutex);
        Triggering_cond.wait(waitlock);
      }

      // We're hijacking NSS for SNEWS trigger
      nova::NovaSNEWSInterface::SNEWSMessageInfo SNEWSinfo;

      nsimessages::SNEWSMessage SNEWSMessage;

      // Listen for message
      requestReceiver.receiveMessage(SNEWSMessage);

      // copy the message data to a SNEWSinfo object
      SNEWSinfo.set(SNEWSMessage.time,SNEWSMessage.type);

      //TODO Need to encode location so we can set the subtype
      // Get which type of trigger it is
      trigID spillType = TRIG_ID_NULL;
      trigSourceSubID whichSignal = TRIGSRC_SUBID_DUMMY;
      if (SNEWSMessage.type==nova::NovaSNEWSInterface::kSNEWSTrigger) {
	spillType = TRIG_ID_SNEWS_TRIGGER;
	whichSignal = TRIGSRC_SUBID_SNEWS_TRIG;
      }
      else if (SNEWSMessage.type==nova::NovaSNEWSInterface::kSNEWSBeatSlow) {
	spillType = TRIG_ID_SNEWS_BEAT_SLOW;
	whichSignal = TRIGSRC_SUBID_SNEWS_BEAT_SLOW;
      }
      else if (SNEWSMessage.type==nova::NovaSNEWSInterface::kSNEWSBeatFast) {
	spillType = TRIG_ID_SNEWS_BEAT_FAST;
	whichSignal = TRIGSRC_SUBID_SNEWS_BEAT_FAST;
      }
      if (spillType==TRIG_ID_NULL) {
	 ::mf::LogWarning("Unknown SNEWS Trigger") << "SNEWSserver has sent an unknown SNEWS type, ID " << SNEWSMessage.type;
	continue;  //bail
      }
      GTInfo *info = &GT_Stats::getInstance().Triggers.at(spillType);

      // If that trigger isn't enabled skip to wait for next one
      if (!info->getEnabled()) continue;

      // Apply prescale
      if (!(info->incSeenCount() % info->getPrescale())) {
        // Get the time out, convert back to t_time
        uint64_t spillNova = SNEWSinfo.getTime();
        timeval spillTv;
        novadaq::timeutils::convertNovaTimeToUnixTime(spillNova,spillTv);
        time_t trigTime = spillTv.tv_sec;

        // report the SNEWS Trigger time
        char ctimestr[27];ctime_r(&trigTime,ctimestr);
	if (!(info->getSeenCount()%logLevel))
	   LOG_DEBUG("Progress") << "Received " << info->getName()
	     << " trigger at time "
             << ctimestr << " with NovaTS=0x"<<hex<<spillNova<<dec;

        // create a trigger atom
	// Version 2 is after 8/12/11's fleshing out of TriggerDefines.h
        RawTrigger trigger((version_t) 2);

        // Next, we construct the different segments of the trigger atom
        // one at a time and stuff them into the atom

        // Start with header
        RawTriggerHeader triggerHeader;
      
        // Set HW trigger flag
        triggerHeader.setHWFlag(0); // calling SNEWS a SW trigger

        // Set source
        triggerHeader.setSource(TRIGSRC_SNEWS);
	//TODO Need to add SourceID for BNL and INFN
        triggerHeader.setSourceID(TRIGSRC_ID_SNEWS_BNL);

	// Which sort of trigger?
        triggerHeader.setSourceSubID(whichSignal);
 
        // Set trigger time.  
        RawTriggerTime triggerTime;

        // This is the time the trigger was generated
        timeval now;
        gettimeofday(&now,0);
        triggerTime.setTriggerTimeStampTv(now);
        uint64_t nova_now;
        novadaq::timeutils::convertUnixTimeToNovaTime(now,nova_now);
        triggerTime.setTriggerTimeStamp(nova_now);

        //TODO Need a SNEWS clocksource
        triggerTime.setClockSource(CLOCKSRC_UNKNOWN);

        // Issue a warning if the spill is "too old".  Manually defined as
        // off from current time by more than 10 minutes, or in the future.
        // Will change this to something configurable later.
        uint64_t deltat = nova_now - spillNova;
	double deltat_s = (double)deltat/(double)timeutils::NOVA_TIME_FACTOR;
	GT_Stats::getInstance().logSNEWSDelay(deltat_s);
        if (spillNova < nova_now) {
	  if (deltat_s > 600)
	    ::mf::LogWarning("Time Shear") << "Spill time is much older than current time " << deltat_s <<" s";
        } else {
          ::mf::LogWarning("Time Shear") << "Spill time is in the future " << deltat_s*(-1000.0) << " ms";
        }

        // pack our time into the trigger atom
        trigger.setTriggerTime(triggerTime);

        // Set trigger time marker
        // This is the actual spill time used to go searching for data
        RawTriggerTimingMarker triggerTimingMarker;
        uint64_t trigTimeNova=spillNova;
        trigTimeNova+=info->getDelay();

        triggerTimingMarker.setTriggerStart_baseClock(trigTimeNova);    
        // pack our time marker into the trigger atom
        trigger.setTriggerTimingMarker(triggerTimingMarker);

        // Set trigger mask
        RawTriggerMask triggerMask;
        triggerMask.setMonteCarloFlag(0);  // Not MC
        triggerMask.setTrigMod(0);         // This is the "Trigger Set"

        // confusingly named method, same as the parent's method that
        // sets this whole thing.  But, this takes a trigger bit number,
        // (trigType-1 here) and sets all the appropriate higher bits if
        // needed
        triggerMask.setTriggerMask(info->getType()-1);

        // set prescale factor
        triggerMask.setPrescale(info->getPrescale());

        // pack our mask into the trigger atom
        trigger.setTriggerMask(triggerMask);

        // let LT calculate info needed to chain the triggers        
        GTLongTrigs LT(info->getLength());
        // let LT adjust trigger times, loop over it and send while it
        // works
        bool first = true;
        while (LT.adjustTrigger(&trigger) > 0) {

	  // Set trigger number.
	  uint64_t TriggerNumberCopy = GT_Stats::getInstance().incTrigNo();
	  // keep track that we're sending a trigger of this type
	  info->incSentCount();
	  triggerHeader.setTriggerNumber(TriggerNumberCopy);
	  if (first) {
	    first = false;
            // Master Trigger number starts off the same
            triggerHeader.setMasterTriggerNumber(TriggerNumberCopy);
	  }
	  // pack our header into the trigger atom - rest was loaded before
	  trigger.setTriggerHeader(triggerHeader);

	  // create the message, pack in the trigger atom, and send it
	  triggermessages::TriggerMessage theMessage;
	  theMessage.payload.length(4 * trigger.sizeofdata());
	  trigger.writeData(theMessage.payload.get_buffer());
	  // People are annoyed by this.  Rather than simply turning off
	  // debug messages, we've made logLevel into a message prescale.

	  if (!(info->getSeenCount()%logLevel))
            LOG_DEBUG("Progress") << "Sending " << info->getName() 
	      << " SNEWS trigger number " << TriggerNumberCopy 
//              <<" with NovaTS=0x"<<hex
//              <<triggerTimingMarker.getTriggerStart_baseClock()
//              <<" with GenTime(Nova)=0x"
//              <<trigger.getTriggerTime()->getTriggerTimeStamp()
//              << dec
//              <<" (gen timeval = "<<now.tv_sec
//              <<" sec, "<<now.tv_usec
//              <<" usec.)"
              <<endl;
	  
	  // Queue the Message to the central sender queues
	  theQueues.QueueTrigger("SNEWS",theMessage);

        }
      } // endif prescale
    } 
}
