#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>

#include <NovaGlobalTrigger/GT_Inst.h>

#include <messagefacility/MessageLogger/MessageLogger.h>

using namespace std;
using namespace novadaq::gt;

// do the unix fork/exec dance to spin off a NovaGlobalTrigger program
// on the specified partition

/**
 * Constructor initializes array of pid's, config files
 *
 */
GT_Inst::GT_Inst() {
  memset(GTProcs,0,NUMPARTS*sizeof(pid_t));
  for (int32_t i=0; i<NUMPARTS; i++) {
    configURI[i] = "";
  }
}


/**
 * Returns the single instance of the GT_Inst.
 *
 * @returns GT_Inst instance.
 */
GT_Inst& GT_Inst::getInstance() {
    static GT_Inst singletonInstance;
    return singletonInstance;
}


// Destructor not needed?
//GT_Inst::~GT_Inst() {
// nothing at present, we dynamically allocate no memory in here to free
//}


/**
 * Returns the pid of the NovaGlobalTrigger instance on the specified partition
 *
 * @returns pid_t or -1 for no instance available
 */
pid_t GT_Inst::GetPID(int32_t partno) const
{
  if (inrange(partno)) {
// would like to mutex protect this read just in case, but it barfs here?
//    boost::mutex::scoped_lock lock(GTProcs_mutex);
    return GTProcs[partno];
  } else {
    return -1;
  }
  // note there is a corner case, pid=0.  But on a linux system, that
  // pid is always the init process, so shouldn't be a problem
}


/**
 * clears the partition table of the pid of a child instance
 * returns the partition on which it was running
 *
 * Only called by the SIGCHLD handler when a partition dies or is killed
 */
int32_t GT_Inst::ClearPID(pid_t pid)
{
// should we check for the pid's existance, in case this gets called
// from somewhere unusual?
  boost::mutex::scoped_lock lock(GTProcs_mutex);
  for (int32_t i=0; i<NUMPARTS; i++) {
    if (GTProcs[i] == pid) {
      GTProcs[i] = 0;
      return i;
    }
  }
  // If we fell through to here, that pid wasn't found
  ::mf::LogWarning("State") << "Couldn't find a partition for pid " << pid
			    << " which we think just died" << endl;
  return -1;
}


/**
 * tries to make a NovaGlobalTrigger on the specified partition
 *
 * @returns 0 for success, 1 for error
 */
int32_t GT_Inst::spawnGT(int32_t partno) 
{
  mf::SetModuleName("spawnGT");

  // range check
  if (!inrange(partno)) {
    ::mf::LogWarning("State") << "Requested partition " 
			      << partno << " is out of range" << endl;
    return 1;
  }

  boost::mutex::scoped_lock lock(GT_Inst_mutex);

  // do we already have an instance running on this partition?
  if (GetPID(partno) > 0) {
    // room for more checking here, could go looking for the pid in 
    // the process table to see if it's still there
    ::mf::LogWarning("State") << "Already a GT running on partition " 
			      << partno << " with pid " 
			      << GTProcs[partno] << endl;
    return 1;
  }

  // Still here?  Go spin it off
  char partStr[(sizeof(int32_t)*8+1)];
  sprintf(partStr,"%d",partno);

  pid_t pID = fork();
  if (pID == 0)                // child
    {
      // Code only executed by child process

      // exec the NovaGlobalTrigger which is first in the path,
      // on the specified partition.
      // ConfigurationManaget can passes us a full config file,
      // so point at the config file with -c 
      execlp("NovaGlobalTrigger", "NovaGlobalTrigger", "-p", 
	     partStr, "-c", configURI[partno].c_str(), (char *) 0);
    }
  else if (pID < 0)            // failed to fork
    {
      ::mf::LogWarning("State") << "Failed to fork GT process" << endl;
      return 1;
    }
  else                                   // parent
    {
      // Code only executed by parent process
      ::mf::LogInfo("Progress") << "NovaGlobalTrigger process " << pID 
				<< " forked on partition " << partno << endl;
    }
  boost::mutex::scoped_lock ProcLock(GTProcs_mutex);
  GTProcs[partno] = pID;
  return 0;
}


/**
 * tries to kill a NovaGlobalTrigger on the specified partition
 *
 * @returns 0 for success, 1 for error
 */
int32_t GT_Inst::killGT(int32_t partno) 
{
  mf::SetModuleName("killGT");

  // range check
  if (!inrange(partno)) {
    ::mf::LogWarning("State") << "Requested partition " 
			      << partno << " is out of range" << endl;
    return 1;
  }

  boost::mutex::scoped_lock lock(GT_Inst_mutex);

  // do we already have an instance running on this partition?
  pid_t pID = GetPID(partno);
  if (pID > 0) {
    // room for more checking here, could go looking for the pid in 
    // the process table to see if it's still there
    char killStr[16];  // should we bounds check this?
    sprintf(killStr,"kill %d",pID);
    int32_t retval = system(killStr);
    if (!retval) {
// let this get reset from the signal handler instead
//      GTProcs[partno] = 0;
      ::mf::LogInfo("Progress") << "GT process " << pID 
				<< " killed on partition " << partno << endl;
      return 0;
    } else {
      ::mf::LogWarning("Progress") << "GT process " << pID 
				   << " NOT killed on partition " 
				   << partno << " with error code " 
				   << retval<< endl;
      return retval;
    }
  } else {
    ::mf::LogWarning("State") << "No GT running on partition " 
			      << partno << " so it can't be killed"  << endl;
    return 1;
  }
}


/**
 * set the config file for the specified partition
 *
 */
bool GT_Inst::setConfigURI(std::string URI, int32_t partno)
{
  // range check
  if (!inrange(partno)) {
    ::mf::LogWarning("State") << "Requested partition " 
			      << partno << " is out of range" << endl;
    return false;
  }

  boost::mutex::scoped_lock lock(GT_Inst_mutex);  

  configURI[partno] = URI;
  return true;
}


/**
 * get the config file for the specified partition
 *
 */
std::string GT_Inst::getConfigURI(int32_t partno)
{
  // range check
  if (!inrange(partno)) {
    ::mf::LogWarning("State") << "Requested partition " 
			      << partno << " is out of range" << endl;
    return "";
  }

  boost::mutex::scoped_lock lock(GT_Inst_mutex);  

  return configURI[partno];
}


/**
 * checks that the specified partition is in the valid range
 *
 * @returns true if in range, false if not
 */
bool GT_Inst::inrange(int32_t partno) const
{
  if ((partno >= 0) && (partno < NUMPARTS)) {
    return true;
  } else {
    return false;  // partition out of range!
  }
}
