#include <map>
#include <messagefacility/MessageLogger/MessageLogger.h>
#include <boost/format.hpp>

#include <NovaGlobalTrigger/GT_SendQueues.h>

GT_NAMESPACE_BEGIN

GT_SendQueues& GT_SendQueues::getInstance(){
    static GT_SendQueues singletonInstance;
    return singletonInstance;
}

bool GT_SendQueues::QueueTrigger(std::string QueueName, 
				 triggermessages::TriggerMessage& msg){

  // First find the pointer to the actual queue that was requested
  std::map<std::string,messageQueue*>::iterator it;
  messageQueue* Qptr;

  // First find the pointer to the messageQueue
  try{
    Qptr = theQueueMap.at(QueueName);
    if(_debug){
      LOG_DEBUG("GT_SendQueues") << "Found Queue" << QueueName;
    }
  }
  catch( const std::out_of_range& oor){
    ::mf::LogWarning("GT_SendQueues") << QueueName 
				      << " not found: Out of range error: " 
				      << oor.what();
    return false;
  }

  // Push the message into the queue
  if(Qptr->queueMessage(msg)){
    if(_debug){
      LOG_DEBUG("GT_SendQueues") << "Message queued.  Current Queue Size " << Qptr->size();
    }
    return true;
  }else{
    ::mf::LogWarning("GT_SendQueues") << "Unable to queue trigger message in: " << QueueName;
    return false;
  }

  return false;
}

bool GT_SendQueues::AddQueue(std::string QueueName, 
			     int priority){
  // Allocate the new messageQueue
  messageQueue* theNewQueue = new messageQueue();
 
 // Put the pointer to the queue into the list by name
  std::pair<std::map<std::string,messageQueue*>::iterator,bool> retVal1;
  std::pair<std::map<std::string,int>::iterator,bool>           retVal2;
  std::pair<std::map<int,messageQueue*>::iterator,bool>         retVal3;

  retVal1 = theQueueMap.insert(std::pair<std::string, messageQueue*>(QueueName, theNewQueue));
  if(retVal1.second == false){
    ::mf::LogWarning("GT_SendQueues") 
      << "Unable to create new Queue. Queue name already exists: " 
      << retVal1.first->second;
    delete theNewQueue;
    return false;
  }

  // Put the name of the queue into the list by priority
  retVal2 = theQueuePriority.insert(std::pair<std::string,int>(QueueName,priority));
  if(retVal2.second == false){
    ::mf::LogWarning("GT_SendQueues") 
      << "Unable to create new Queue with prioirty: " << priority 
      << ". Queue name already exists: " 
      << retVal2.first->second;
 
    // Remove the previous entry that was inserted
    theQueueMap.erase(QueueName);
    delete theNewQueue;
    return false;
  }

  // Put the pointer to the queue into the list by priority
  retVal3 = theAccessQueues.insert(std::pair<int,messageQueue*>(priority,theNewQueue));
  if(retVal3.second == false){
    ::mf::LogWarning("GT_SendQueues") 
      << "Unable to create new priority for Queue"
      << "Priority value: " << priority << "Already already exists: " 
      << retVal3.first->second;
 
    // Remove the previous entry that was inserted
    theQueueMap.erase(QueueName);
    theQueuePriority.erase(QueueName);
    delete theNewQueue;
    return false;
  }

  return true;
}

bool GT_SendQueues::RemoveQueue(std::string QueueName){
  std::map<std::string,messageQueue*>::iterator it;
  std::map<std::string,int>::iterator           it2;
  std::map<int, messageQueue*>::iterator        it3;


  messageQueue* Qptr;
  int priorityVal;

 
  // First find the pointer to the messageQueue
  try{
    Qptr = theQueueMap.at(QueueName);
  }
  catch( const std::out_of_range& oor){
    ::mf::LogWarning("GT_SendQueues") << QueueName 
				      << " not found: Out of range error: " 
				      << oor.what();
    return false;
  }

  // Then find its priority Index
  try{
    priorityVal = theQueuePriority.at(QueueName);
  }
  catch( const std::out_of_range& oor){
    ::mf::LogWarning("GT_SendQueues") << QueueName 
				      << " not found: Out of range error: " 
				      << oor.what();
    return false;
  }

  // From these get the iterators to the positions in the maps

  // Look for the QueueName
  it=theQueueMap.find(QueueName);
  
  // If it was found then remove it from the list
  if(it != theQueueMap.end()){
    theQueueMap.erase(it);
  }

  // Do the same for the QueueName in the priorty map
  it2=theQueuePriority.find(QueueName);
  if(it2 != theQueuePriority.end()){
    theQueuePriority.erase(it2);
  }

  // Do the same for the priority indexed map
  it3=theAccessQueues.find(priorityVal);
  if(it3 != theAccessQueues.end()){
    theAccessQueues.erase(it3);
  }

  return true;
}


void GT_SendQueues::ChangePriority(std::string QueueName, 
				   int priority){
  int oldPriorityVal=-1;
  messageQueue* Qptr;

  // This is tricky.  First we get the old priority
  try{
    oldPriorityVal = theQueuePriority.at(QueueName);
  }
  catch( const std::out_of_range& oor){
    ::mf::LogWarning("GT_SendQueues") << QueueName 
				      << " not found: Out of range error: " 
				      << oor.what();
    return;
  }

  // And the pointer to the Queue
  // First find the pointer to the messageQueue
  try{
    Qptr = theQueueMap.at(QueueName);
  }
  catch( const std::out_of_range& oor){
    ::mf::LogWarning("GT_SendQueues") << QueueName 
				      << " not found: Out of range error: " 
				      << oor.what();
  }

  std::map<int, messageQueue*>::iterator        it3;
  
  it3=theAccessQueues.find(oldPriorityVal);
  if(it3 != theAccessQueues.end()){
    theAccessQueues.erase(it3);
  }

  // Then make a new priority key'd entry
  // Put the pointer to the queue into the list by priority
  std::pair<std::map<int,messageQueue*>::iterator,bool>         retVal3;
  retVal3 = theAccessQueues.insert(std::pair<int,messageQueue*>(priority,Qptr));

  // Then update the priority in the name indexed map
  theQueuePriority.at(QueueName) = priority;

}

int GT_SendQueues::GetPriority(std::string QueueName){
  int priorityVal=-1;
  try{
    priorityVal = theQueuePriority.at(QueueName);
  }
  catch( const std::out_of_range& oor){
    ::mf::LogWarning("GT_SendQueues") << QueueName 
				      << " not found: Out of range error: " 
				      << oor.what();
  }
  
  return priorityVal;
}


void GT_SendQueues::ListQueues(){
  std::map<std::string,messageQueue*>::iterator it;
  for(it=theQueueMap.begin(); it!=theQueueMap.end(); ++it){
    ::mf::LogInfo("ListQueues")  
      << "Queue Name: " << it->first << ", at address "
      << boost::format("0x%p")%it->second;
  } // endfor
}

triggermessages::TriggerMessage* GT_SendQueues::GetNextTrigger(){
  int retry=0;
  static triggermessages::TriggerMessage* nillMessage = NULL;

  static size_t max_sizes[10]={0,0,0,0,0,0,0,0,0,0};
  static int reportIdx=0;

  std::map<int,messageQueue*>::iterator it;

  int max_retry = _pollTimeOut*1000000/_pollVal;

  //LOG_DEBUG("GT_SendQueues") << "Max Retry: " << max_retry 
  // 	    << " PollVal: " << _pollVal 
  //	    << " PollTimeOut: " << _pollTimeOut ;

  while( retry < max_retry){
    if(!(retry%1000)){
      //      LOG_DEBUG("GT_SendQueues") << "Retry: " << retry << std::endl;
    }
    // Cycle over the list of available trigger queues
    int i=0;
      ++reportIdx;
      
    for(it=theAccessQueues.begin(); it!=theAccessQueues.end(); ++it){
      // Check if any of them have a non-zero size (triggers waiting)

      if(it->second->size() > max_sizes[i]){
	max_sizes[i] = it->second->size();
      }

      if(_debug){
	if(!(reportIdx%1000)){
	  LOG_DEBUG("GT_SendQueues") << "Queue: "     << it->first 
		    << " Max Size: " << max_sizes[i]
				     << " Current Size: " 
				     << it->second->size();
	}
      }
      ++i;
      if(it->second->size() > 0){
	//LOG_DEBUG("GT_SendQueues") << "Popping Queue" << "..." ;

	// Pop off the message
	// triggermessages::TriggerMessage& tr= it->second->popMessage();
	//LOG_DEBUG("GT_SendQueues") << "Popped!" << std::endl;
	// return tr;
	return it->second->popMessage();
      }    
    }
    usleep(_pollVal);
    ++retry;
    
  } // endwhile
  retry = 0;
  throw NextTrigger_TimeOut;
  return nillMessage;
}

GT_NAMESPACE_END
