#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <string>
#include <sstream>

#include <messagefacility/MessageLogger/MessageLogger.h>

#include <DAQDataFormats/TriggerDefines.h>

#include <NovaGlobalTrigger/GT_Stats.h>
#include <NovaGlobalTrigger/GTGlobals.h>

using namespace std;
using namespace novadaq::gt;
using namespace daqdataformats;

// Wrangle trigger numbers and other such counting statistics

/**
 * Constructor sets things to zero, sets up ganglia metrics
 *
 */
GT_Stats::GT_Stats() {
  // Make an array of GTInfo objects to keep track of triggers
  // This takes the vector declared in the header but instantiates 
  // all the elements.  By using resize() we actually take the
  // predeclared Triggers and grow it.  Can't do that in the .h.  Can't 
  // just declare it here like Triggers(MAX_TRIGGER_BITS+1) or that makes
  // a new copy.  Dunno why, but that's a couple hours of my life I'll
  // never get back.

  Triggers.resize(MAX_TRIGGER_BITS+1);  // there is one more type than bits
                                        // since we need TRIG_ID_NULL
  resetTrigNo();

  // Create custom metrics for monitoring trigger rates
  int32_t partno;
  {
    boost::mutex::scoped_lock lock(partNumber_mutex);
    partno = partNumber;
  }
  // set ganglia partition.
  _daqMonClient.setPartition(partno);
  // What do do about group?  Just make one.  I think if we want
  // different groups coming from the same apps, trouble happens
  _daqMonClient.setGroupName("GT");
  TriggerRate = _daqMonClient.createMetric<double>
    ("Overall Trigger Rate", "Hz", nova::novadaqmonitorclient::RATE, "",
     "Combined rate of all triggers");
}


/**
 * Returns the single instance of the GT_Stats.
 *
 * @returns GT_Stats instance.
 */
GT_Stats& GT_Stats::getInstance() {
    static GT_Stats singletonInstance;
    return singletonInstance;
}


// Destructor not needed?
//GT_Stats::~GT_Stats() {
// nothing at present, we dynamically allocate no memory in here to free
//}


/**
 * Returns the current trigger number
 *
 */
uint64_t GT_Stats::getTrigNo()
{
  boost::mutex::scoped_lock lock(TriggerNumber_mutex);
  return TriggerNumber;
}

/**
 * returns ++TriggerNumber and tickles ganglia
 *
 */
uint64_t GT_Stats::incTrigNo()
{
  boost::mutex::scoped_lock lock(TriggerNumber_mutex);
  _daqMonClient.setValue(TriggerRate,double(1.));
  return ++TriggerNumber;
}


/**
 * zeros trigger number and all the other stats
 *
 */
void GT_Stats::resetTrigNo() 
{
  boost::mutex::scoped_lock lock(TriggerNumber_mutex);
  TriggerNumber       = 0;

  // Loop over the counts in the Triggers array
  for (size_t i=0; i<(MAX_TRIGGER_BITS+1); i++) {
    Triggers.at(i).resetSeenCount();
    Triggers.at(i).resetSentCount();
    Triggers.at(i).resetMissedCount();
    if (Triggers.at(i).Queue != NULL) Triggers.at(i).Queue->flush();
  } 
}


/**
 * returns trigger count by trigger bit, in array, returning trigger number
 *
 */
uint64_t GT_Stats::reportTrigNo(uint64_t counts[], uint32_t numcounts)
{
  memset(counts,0,numcounts*sizeof(uint64_t));

  boost::mutex::scoped_lock lock(TriggerNumber_mutex);

  std::stringstream outstring;
  outstring << "There were " << TriggerNumber
            << " total triggers,\n";
  for (uint32_t i=0; i<numcounts; i++) {
    // Triggers dumped into array by bit number, see
    // DAQDataFormats/TriggerDefines.* 
    counts[i] = Triggers.at(i+1).getSentCount();
    if (counts[i]>0)
      outstring << counts[i] << " "
		<< getTriggerName(Triggers.at(i+1).getType())
	        << endl;
  }
  ::mf::LogInfo("Progress") << outstring.str();

  return TriggerNumber;
}


/**
 * create ganglia spill metrics
 *
 */
void GT_Stats::initSpillMetrics()
{
  spillTriggerDelayMAX = _daqMonClient.createMetric<double>
    ("Spill Trigger Max Delay","ms", nova::novadaqmonitorclient::MAXIMUM,
     "", "All spill triggers delay, plotting maximums");
  spillTriggerDelayMIN = _daqMonClient.createMetric<double>
    ("Spill Trigger Min Delay","ms", nova::novadaqmonitorclient::MINIMUM,
     "", "All spill triggers delay, plotting minimums");
  spillTriggerDelayAVG = _daqMonClient.createMetric<double>
    ("Spill Trigger Avg Delay","ms", nova::novadaqmonitorclient::AVERAGE,
     "", "All spill triggers delay, plotting the average");
  spillTriggerDelaySTD = _daqMonClient.createMetric<double>
    ("Spill Trigger Std Delay","ms", nova::novadaqmonitorclient::STDEV,
     "", "All spill triggers delay, plotting the standard deviation");
}


/**
 * create ganglia SNEWS metrics
 *
 */
void GT_Stats::initSNEWSMetrics()
{
  snewsTriggerDelayMAX = _daqMonClient.createMetric<double>
    ("SNEWS Trigger Max Delay","ms", nova::novadaqmonitorclient::MAXIMUM,
     "", "All SNEWS triggers delay, plotting maximums");
  snewsTriggerDelayMIN = _daqMonClient.createMetric<double>
    ("SNEWS Trigger Min Delay","ms", nova::novadaqmonitorclient::MINIMUM,
     "", "All SNEWS triggers delay, plotting minimums");
  snewsTriggerDelayAVG = _daqMonClient.createMetric<double>
    ("SNEWS Trigger Avg Delay","ms", nova::novadaqmonitorclient::AVERAGE,
     "", "All SNEWS triggers delay, plotting the average");
  snewsTriggerDelaySTD = _daqMonClient.createMetric<double>
    ("SNEWS Trigger Std Delay","ms", nova::novadaqmonitorclient::STDEV,
     "", "All SNEWS triggers delay, plotting the standard deviation");
}


/**
 * send spill trigger delay to ganglia
 *
 */
void GT_Stats::logSpillDelay(double delay)
{
  double msDelay = delay*1000.0;
  _daqMonClient.setValue(spillTriggerDelayMAX,msDelay);
  _daqMonClient.setValue(spillTriggerDelayAVG,msDelay);
  _daqMonClient.setValue(spillTriggerDelayMIN,msDelay);
  _daqMonClient.setValue(spillTriggerDelaySTD,msDelay);
}


/**
 * send SNEWS trigger delay to ganglia
 *
 */
void GT_Stats::logSNEWSDelay(double delay)
{
  double msDelay = delay*1000.0;
  _daqMonClient.setValue(snewsTriggerDelayMAX,msDelay);
  _daqMonClient.setValue(snewsTriggerDelayAVG,msDelay);
  _daqMonClient.setValue(snewsTriggerDelayMIN,msDelay);
  _daqMonClient.setValue(snewsTriggerDelaySTD,msDelay);
}


/**
 * start ganglia heartbeat
 *
 */
void GT_Stats::startHeartbeat()
{
  std::string title ="";
  std::string description="Activity on partition (1=active/0=inactive).";
  _monIdPartitionHeartbeat=_daqMonClient.createMetric<double>("Activity","",nova::novadaqmonitorclient::HEARTBEAT,title,description);
}

/**
 * stop ganglia heartbeat
 *
 */
void GT_Stats::stopHeartbeat()
{
  _daqMonClient.destroyMetric(_monIdPartitionHeartbeat); // silence heartbeat
  _monIdPartitionHeartbeat=-1;
}

// private methods.
