#include <NovaGlobalTrigger/GT_Throttle.h>

#include <sys/types.h>
#include <sys/time.h>
#include <string>
#include <deque>

#include <boost/thread/mutex.hpp>


#include <messagefacility/MessageLogger/MessageLogger.h>

// #include <DAQDataFormats/TriggerDefines.h>

// #include <NovaTimingUtilities/TimingUtilities.h>

#include <DAQMessages/ccpp_GTMessages.h>
#include <DAQMessages/ccpp_DDTMessages.h>
#include <NovaGlobalTrigger/GTGlobals.h>
#include <NovaGlobalTrigger/GTLongTrigs.h>


#include <NovaGlobalTrigger/GT.h>
#include <NovaGlobalTrigger/GTInfo.h>

// GT_NAMESPACE_BEGIN

throttlePipeline::throttlePipeline():maxRate(-1),averageRate(0.0),
				     rateSum(0),max_queue_depth(-1),_debug(false){
}

size_t throttlePipeline::size(){
  return theQueue.size();  
}

bool throttlePipeline::empty(){
  return theQueue.empty(); 
};

void throttlePipeline::flush(){
  theQueue.clear();
}

void throttlePipeline::reset(){
  theQueue.clear();
  maxRate          = -1.0;
  averageRate      =  0.0;
  rateSum          =  0.0;
  max_queue_depth  = -1.0;
}

bool throttlePipeline::throttleState(){
  if(_debug){
    LOG_DEBUG("GT_Throttle") << "************ Average Rate: " << averageRate << " Max Rate " << maxRate << " size " << theQueue.size() << " depth " << max_queue_depth;
  }
  if(maxRate < 0.0 ){return false;}         // Throttle disabled
  if(theQueue.size() == 1){ return false;}  // If we have only a single entry we don't trigger a throttle
  if ((novadaq::gt::saneStats) && ((unsigned)max_queue_depth > theQueue.size())){ return false;} // if saneStats set and queue not full, don't throttle

  return (averageRate > maxRate) ? true : false ; // Check if we should throttle or not
}


bool throttlePipeline::pushFront(timeval evtTime){


  // If depth is set negative (unlimited)
  if(max_queue_depth < 0){
    // Push the message to the back of the queue
    theQueue.push_front(evtTime);
    updateAverage();

    return true;
  }
  
  // If the queue depth is non-negative (fixed size)
  if((unsigned)max_queue_depth <= theQueue.size()){
    // Queue is full (so add one element on and drop one element off)

    // Push the new message on the front
    theQueue.push_front(evtTime);      

    // Drop the oldest element from the back
    theQueue.pop_back();

    updateAverage();

    return true;
  }else{
    // Queue is not yet full (so just add an element on to it)
    // Push the message to the back of the queue
    theQueue.push_front(evtTime);
    updateAverage();
    return true;
  }
  
  return false;
}

double throttlePipeline::updateAverage(){
    // Values for computing the average rate
    timeval deltaT;
    timeval f = theQueue.front();
    timeval b = theQueue.back();

    if( timercmp(&f,&b, > )){
      timersub(&f,&b,&deltaT);
    }else{
      timersub(&b,&f,&deltaT);
    };

    // Recompute the running average
    rateSum = ( deltaT.tv_sec * 1000000.0 + 1.0*deltaT.tv_usec )/1000000;
    averageRate = (1.0*theQueue.size())/rateSum;
    if(_debug){
      LOG_DEBUG("GT_Throttle") << "************ Rate Sum: " << rateSum << "  Average Rate: " << averageRate;
    }

    return averageRate;
}

int throttlePipeline::setMaxQueueDepth(int value){
  max_queue_depth = value;
  return max_queue_depth;
}

int throttlePipeline::getMaxQueueDepth(){
  return max_queue_depth;
}

void throttlePipeline::popFront(){
  theQueue.pop_front();
}

// GT_NAMESPACE_END
