/**
 * The RC comms methods for the NGTMaster process.
 * Only deal with partition making.
 * The NovaGlobalTrigger processes do begin, end runs etc.
 */

#include <string>
#include <iostream>
#include <fstream>

#include <rms/base/RmsDestination.h>
#include <NovaDAQUtilities/LocationUtils.h>
#include <NovaGlobalTrigger/NGTMasterRC.h>
#include <NovaGlobalTrigger/GT_Inst.h>

using namespace gov::fnal::cd::rms;

using namespace std;

GT_NAMESPACE_BEGIN

/**
 * Constructor: Creates an instance with the specified application name.
 */
NGTMasterRC::NGTMasterRC(std::string applicationName, int partition) : 
  _applicationName(applicationName) {

  // message facility already started my main program, just SetModuleName
  mf::SetModuleName("NGTMasterRC");

  std::vector<std::string> bcastTargets;
  bcastTargets.push_back(base::RmsDestination::ALL_ELEMENT_TARGET);

  _rcReceiver.reset(new rcclient::RunControlReceiver(applicationName, bcastTargets));

  if ( !changeState(PARTITION_ESTABLISHED,"",partition) ) {
    ::mf::LogError("State") << "Unable to establish new partition number " 
			    << partition << endl;
    abort();
  } else {
    // add listeners for the messages we can now do
    _rcReceiver->addListener<runcontrolmessages::LoadRunConfigurationRequest,
       NGTMasterRC, runcontrolmessages::LoadRunConfigurationResponse>
       (partition, this);

    _rcReceiver->addListener<runcontrolmessages::ConfigureRunRequest,
       NGTMasterRC, runcontrolmessages::ConfigureRunResponse>
       (partition, this);

    _rcReceiver->addListener<runcontrolmessages::StatusRequest,
       NGTMasterRC, runcontrolmessages::StatusResponse>
       (partition, this);

  }
  
  ::mf::LogInfo("Progress") << "NGTMasterRC thread started" << endl;
}


/**
 * Destructor: clean up anything that we have allocated.
 */
NGTMasterRC::~NGTMasterRC() {
  if (_rcReceiver.get() != 0) {
    _rcReceiver->disconnectFromPartition(base::RmsDestination::NULL_PARTITION);
    _rcReceiver.reset();
  }
}


/**
 * Thread run method.
 */
void NGTMasterRC::run() {
  _running = true;

  // not much to do - the message handler is taking care of things
  while (! _stopRequested) {
    sleep(5);
  }

  _running = false;
}


/**
 * Actually establish the partition
 */
bool NGTMasterRC::_establishPartition(int32_t partition) 
{
// process is now spun off in the Configure step not here, so this
// is just a placeholding function for this step in state.
  return true;
  // Start the partition
//      int procStatus = novadaq::gt::GT_Inst::getInstance().spawnGT(partition);
//      if (procStatus == 0)
//        return true;
//      else  
//        return false;
};


/**
 * Actually release the partition
 */
bool NGTMasterRC::_detachFromPartition(int32_t partition) 
{
  // Stop the partition
// this is now handled by the _deConfigureRun
//      int procStatus = novadaq::gt::GT_Inst::getInstance().killGT(partition);
//
//      if (procStatus == 0) 
//        return true;
//      else
//        return false;
    _rcReceiver->disconnectFromPartition(partition);
  return true;
}


/**
 * Handles load run configuration messages.
 */
boost::shared_ptr<runcontrolmessages::LoadRunConfigurationResponse> NGTMasterRC::
  handleMessage(boost::shared_ptr<runcontrolmessages::LoadRunConfigurationRequest> message)
{
  boost::mutex::scoped_lock sl(_callbackMutex);

  // initialize the status that we'll send in the reply to success
  rmscore::RequestStatus reqStatus;
  reqStatus.code = 0;
  reqStatus.message = DDS::string_dup("Success");

  // parse the request
  int partitionNumber = message->header.partitionNumber;
  std::string configURI(message->configurationDataURL);
  ::mf::LogInfo("State") << "LoadRunConfigurationRequest received for "
                   << "partition number " << partitionNumber 
	           << " URI " << configURI << ".";

  // Change the state
  if ( !changeState(RUN_CONFIGURATION_LOADED,configURI,partitionNumber) ) {
    reqStatus.code = 1;
    std::string failString = "Failed to load run config with " + configURI + ".";
    reqStatus.message = DDS::string_dup(failString.c_str());
  }

  // create the reply
  boost::shared_ptr<runcontrolmessages::LoadRunConfigurationResponse> transResponse;
  if (reqStatus.code >= -100) {
    transResponse.reset(new runcontrolmessages::LoadRunConfigurationResponse());
    transResponse->status = reqStatus;
  }
  return transResponse;
}


/**
 * Actually load the run configuration
 */
bool NGTMasterRC::_loadRunConfiguration ( std::string configURI, int32_t partition)
{
  // Here, "load" means set up the string with the config file in GT_Inst

  // is it a valid file
  std::string fname = LocationUtils::translateURL(configURI);
  ifstream configFile;
  configFile.open(fname.c_str());
  if (configFile.is_open()) {
    configFile.close();
    if (novadaq::gt::GT_Inst::getInstance().setConfigURI(configURI,partition))
      return true;
    else {
      ::mf::LogWarning("State") << "Unable to load config file " << configURI
	                  << " on partition " << partition << endl;
      return false;
    }
  } else {
    ::mf::LogWarning("State") << "Unable to open config file " << configURI
                        << endl;
	return false;
  }
}


/**
 * Handles heartbeat messages
 */
boost::shared_ptr<runcontrolmessages::StatusResponse> NGTMasterRC::
  handleMessage(boost::shared_ptr<runcontrolmessages::StatusRequest> message)
{
  boost::mutex::scoped_lock sl(_callbackMutex);

  // initialize the status that we'll send in the reply to success
  rmscore::RequestStatus reqStatus;
  reqStatus.code = 0;
  reqStatus.message = DDS::string_dup("Success");

  // parse the request
  //int partitionNumber = message->header.partitionNumber;
  // ::mf::LogInfo("State") << "StatusRequest received for "
  //                 << "partition number " << partitionNumber 
  //                 << ".";

  // create the reply
  boost::shared_ptr<runcontrolmessages::StatusResponse> transResponse;
  if (reqStatus.code >= -100) {
    transResponse.reset(new runcontrolmessages::StatusResponse());
    transResponse->status = reqStatus;
  }
  return transResponse;
}


/**
 * Handles run configuration messages.
 */
boost::shared_ptr<runcontrolmessages::ConfigureRunResponse> NGTMasterRC::
  handleMessage(boost::shared_ptr<runcontrolmessages::ConfigureRunRequest> message)
{
  boost::mutex::scoped_lock sl(_callbackMutex);

  // initialize the status that we'll send in the reply to sucess
  rmscore::RequestStatus reqStatus;
  reqStatus.code = 0;
  reqStatus.message = DDS::string_dup("Success");

  // parse the request
  int partitionNumber = message->header.partitionNumber;
  ::mf::LogInfo("State") << "ConfigureRunRequest received for "
                   << "partition number " << partitionNumber << ".";

  // Change the state
  if ( !changeState(RUN_CONFIGURED,"",partitionNumber) ) {
    ::mf::LogWarning("State") << "Unable to configure and spawn NovaGlobalTrigger"
	  << " on partition " << partitionNumber << endl;
    reqStatus.code = 1;
    reqStatus.message = DDS::string_dup("Unable to configure and spawn NovaGlobalTrigger");
  }

  // create the reply
  boost::shared_ptr<runcontrolmessages::ConfigureRunResponse> transResponse;
  transResponse.reset(new runcontrolmessages::ConfigureRunResponse());
  transResponse->status = reqStatus;
  return transResponse;
}


/**
 * Configure the run means spawn the process and go
 */
bool NGTMasterRC::_configureRun(int32_t partition) 
{
  // We should be all configured now, so we can actually fork/exec
  // the NovaGlobalTrigger process now
  int procStatus =
  novadaq::gt::GT_Inst::getInstance().spawnGT(partition);
  if (procStatus == 0)
    return true;
  else
    return false;
};


/**
 * kill the NovaGlobalTrigger process
 */
bool NGTMasterRC::_deconfigureRun(int32_t partition) 
{
  // Stop the partition
  int procStatus = novadaq::gt::GT_Inst::getInstance().killGT(partition);

  if (procStatus == 0) 
    return true;
  else
    return false;
};


/**
 * Clear the run configuration
 */
void NGTMasterRC::_unloadRunConfiguration(int32_t partition) 
{
  novadaq::gt::GT_Inst::getInstance().setConfigURI("",partition);
}

GT_NAMESPACE_END
