#include <NovaRunControlClient/RmsMessageClient.h>

namespace novadaq {

namespace rcclient {

/**
 * Creates a receiver wrapper instance from the specified parameters.
 */
ReceiverWrapper::
ReceiverWrapper(CloseablePtr receiverPtr, const std::string& msgType,
                const std::string& targetName, bool canBeOverridden) :
    _receiverPtr(receiverPtr), _msgType(msgType),
    _targetName(targetName), _canBeOverridden(canBeOverridden)
{}

/**
 * Creates a receiver instance with the specified application name and
 * list of broadcast targets to listen to.
 */
RmsMessageClient::
RmsMessageClient(const std::string& applicationName,
                 std::vector<std::string> bcastTargetList,
                 const std::string& rmsMessageChannel,
                 bool provideDefaultHandlers) :
    _rmsMessageChannel(rmsMessageChannel),
    _provideDefaultHandlers(provideDefaultHandlers) {

    // process input arguments, but protect against duplicates
    _applicationName = applicationName;
    bcastTargetList.erase(std::remove(bcastTargetList.begin(),
                                      bcastTargetList.end(),
                                      applicationName),
                          bcastTargetList.end());
    while (bcastTargetList.size() > 0) {
        std::string targetName = bcastTargetList[0];
        bcastTargetList.erase(std::remove(bcastTargetList.begin(),
                                          bcastTargetList.end(),
                                          targetName),
                              bcastTargetList.end());
        _bcastTargetList.push_back(targetName);
    }

    // check if a debug flag has been specified
    _debugFlag = 0;
    std::string debugString = novadaq::EnvVarCache::getInstance().
        getEnvVar("RMS_MESSAGE_CLIENT_DEBUG");
    if (debugString.length() > 0) {
        try {_debugFlag = boost::lexical_cast<int>(debugString);}
        catch (...) {}
    }

    _workerThread =
        new boost::thread(boost::bind(&novadaq::Runnable::run, this));
}

/**
 * Destructor.
 */
RmsMessageClient::~RmsMessageClient() {

    // stop and clean up our worker thread
    this->stop();
    _workerThread->join();
    delete _workerThread;

    // clean up all receivers in all partitions (if any)
    ReceiverWrapperListMap::iterator rmapIter;
    ReceiverWrapperListMap::iterator rmapEnd = _receiverListTable.end();
    for (rmapIter=_receiverListTable.begin(); rmapIter!=rmapEnd; rmapIter++) {
        ReceiverWrapperListPtr receiverListPtr = rmapIter->second;

        ReceiverWrapperList::iterator rcvrIter;
        ReceiverWrapperList::iterator rcvrEnd = receiverListPtr->end();
        for (rcvrIter=receiverListPtr->begin(); rcvrIter!=rcvrEnd; rcvrIter++) {
            CloseablePtr receiverPtr = (*rcvrIter)->getReceiverPtr();
            receiverPtr->close();
        }

        _receiverListTable.erase(rmapIter);
    }

    // clean up all senders in all partitions (if any)
    CloseableMapMap::iterator smapIter;
    CloseableMapMap::iterator smapEnd = _senderListTable.end();
    for (smapIter=_senderListTable.begin(); smapIter!=smapEnd; smapIter++) {
        CloseableMapPtr senderMapPtr = smapIter->second;

        CloseableMap::iterator sndrIter;
        CloseableMap::iterator sndrEnd = senderMapPtr->end();
        for (sndrIter=senderMapPtr->begin(); sndrIter!=sndrEnd; sndrIter++) {
            CloseablePtr senderPtr = sndrIter->second;
            senderPtr->close();
        }

        _senderListTable.erase(smapIter);
    }

    // clean up any remaining DDS connections
    DDSConnectionMap::iterator connIter;
    DDSConnectionMap::iterator connEnd = _connectionTable.end();
    for (connIter = _connectionTable.begin(); connIter != connEnd; connIter++) {
        DDSConnectionPtr ddsConnection = connIter->second;
        ddsConnection->close();

        _connectionTable.erase(connIter);
    }
}

/**
 * Disconnects the receiver from the specified partition.
 */
void RmsMessageClient::disconnectFromPartition(int partitionNumber) {

    // create the request to disconnect
    boost::shared_ptr<Action>
        disconnectRequest(new PartitionDisconnecter(this, partitionNumber));

    // enqueue the request
    this->enqueueAction(disconnectRequest);
}

/**
 * Enqueues the specified Action for execution.
 */
void RmsMessageClient::enqueueAction(boost::shared_ptr<Action> theAction) {
    _workQueue.enq_wait(theAction);
}

/**
  * Thread run method.
  */
void RmsMessageClient::run() {
    _stopRequested = false;
    _running = true;

    while (! _stopRequested) {
        boost::shared_ptr<Action> nextAction;
        if (_workQueue.deq_timed_wait(nextAction, 1)) {
            if (nextAction.get() != 0) {  // paranoia
                int partNumForAction = nextAction->getPartitionNumber();
                //std::cout << "Valid nextAction ["
                //          << nextAction->getDescription()
                //          << "] for partition "
                //          << partNumForAction
                //          << ", queue size = "
                //          << _workQueue.size() << std::endl;
                //std::cout.flush();

                bool doAction = false;
                {
                    boost::mutex::scoped_lock sl(_workListMutex);
                    DDSConnectionMap::iterator connIter =
                        _connectionTable.find(partNumForAction);
                    if (connIter != _connectionTable.end()) {
                        doAction = true;
                    }
                }

                if (doAction) {
                    nextAction->executeAction();
                }
                //else {
                //    std::cout << "Skipping execution of Action for invalid "
                //              << "partition number " << partNumForAction
                //              << std::endl;
                //    std::cout.flush();
                //}
            }
        }
    }

    _running = false;
}

/**
 * Disconnects the receiver from the specified partition.
 */
void RmsMessageClient::_disconnectFromPartition(int partitionNumber) {
    boost::mutex::scoped_lock sl(_workListMutex);

    // clean up any senders in the partition
    CloseableMapMap::iterator smapIter = _senderListTable.find(partitionNumber);
    if (smapIter != _senderListTable.end()) {
        CloseableMapPtr senderMapPtr = smapIter->second;

        CloseableMap::iterator sndrIter;
        for (sndrIter = senderMapPtr->begin();
             sndrIter != senderMapPtr->end();
             sndrIter++) {
            CloseablePtr senderPtr = sndrIter->second;
            senderPtr->close();
        }

        _senderListTable.erase(smapIter);
    }

    // clean up all receivers in the partition
    ReceiverWrapperListMap::iterator rmapIter =
        _receiverListTable.find(partitionNumber);
    if (rmapIter != _receiverListTable.end()) {
        ReceiverWrapperListPtr receiverListPtr = rmapIter->second;

        ReceiverWrapperList::iterator rcvrIter;
        for (rcvrIter = receiverListPtr->begin();
             rcvrIter != receiverListPtr->end();
             rcvrIter++) {
            CloseablePtr receiverPtr = (*rcvrIter)->getReceiverPtr();
            receiverPtr->close();
        }

        _receiverListTable.erase(rmapIter);
    }

    // clean up the DDS connection, if it exists
    DDSConnectionMap::iterator connIter =
        _connectionTable.find(partitionNumber);
    if (connIter != _connectionTable.end()) {
        DDSConnectionPtr ddsConnection = connIter->second;
        ddsConnection->close();

        _connectionTable.erase(connIter);
    }
}

};  // end of namespace rcclient

};  // end of namespace novadaq
