#include <NovaRunControlClient/RunControlReceiver.h>

namespace novadaq {

namespace rcclient {

boost::mutex RunControlReceiver::_instanceMutex;
std::vector<int64_t> RunControlReceiver::_instanceList;

/**
 * Creates a receiver instance with the specified application name and
 * list of broadcast targets to listen to.
 */
RunControlReceiver::
RunControlReceiver(std::string applicationName,
                   std::vector<std::string> bcastTargetList,
                   bool provideDefaultHandlers) :
    RmsMessageClient(applicationName, bcastTargetList,
                     RMSNS::base::RmsDestination::CONTROL_CHANNEL,
                     provideDefaultHandlers)
{
    // add this instance to this instanceList
    {
        boost::mutex::scoped_lock sl(_instanceMutex);
        _instanceList.push_back((int64_t) this);
    }
}

/**
 * Destructor.
 */
RunControlReceiver::~RunControlReceiver()
{
    // remove this instance from this instanceList
    {
        boost::mutex::scoped_lock sl(_instanceMutex);
        for (uint32_t idx = 0; idx < _instanceList.size(); ++idx) {
            if (_instanceList[idx] == (int64_t) this) {
                _instanceList.erase(_instanceList.begin()+idx);
                break;
            }
        }
    }
}

/**
 * Tests if the current instance should install default handlers.
 */
bool RunControlReceiver::_thisInstanceShouldInstallDefaultHandlers()
{
    boost::mutex::scoped_lock sl(_instanceMutex);
    return (_instanceList.size() > 0 &&
            ((int64_t) this) == _instanceList[0]);
}

/**
 * Adds the full set of default listeners to this receiver instance
 * for the specified destination.
 */
void RunControlReceiver::
_addDefaultListeners(DDSConnectionPtr ddsConnection,
                     RMSNS::base::RmsDestination destination,
                     int partitionNumber,
                     ReceiverWrapperListPtr receiverListPtr) {

    _addOneDefaultListener<runcontrolmessages::VersionInfoRequest,
        runcontrolmessages::VersionInfoResponse>(ddsConnection,
						 destination,
						 partitionNumber,
						 receiverListPtr);
    _addOneDefaultListener<runcontrolmessages::LoadConnectionsConfigurationRequest,
        runcontrolmessages::LoadConnectionsConfigurationResponse>(ddsConnection,
                                                                  destination,
                                                                  partitionNumber,
                                                                  receiverListPtr);
    _addOneDefaultListener<runcontrolmessages::MakeConnectionsRequest,
        runcontrolmessages::MakeConnectionsResponse>(ddsConnection,
                                                     destination,
                                                     partitionNumber,
                                                     receiverListPtr);
    _addOneDefaultListener<runcontrolmessages::LoadHardwareConfigurationRequest,
        runcontrolmessages::LoadHardwareConfigurationResponse>(ddsConnection,
                                                               destination,
                                                               partitionNumber,
                                                               receiverListPtr);
    _addOneDefaultListener<runcontrolmessages::ConfigureHardwareRequest,
        runcontrolmessages::ConfigureHardwareResponse>(ddsConnection,
                                                       destination,
                                                       partitionNumber,
                                                       receiverListPtr);
    _addOneDefaultListener<runcontrolmessages::LoadRunConfigurationRequest,
        runcontrolmessages::LoadRunConfigurationResponse>(ddsConnection,
                                                          destination,
                                                          partitionNumber,
                                                          receiverListPtr);
    _addOneDefaultListener<runcontrolmessages::ConfigureRunRequest,
        runcontrolmessages::ConfigureRunResponse>(ddsConnection,
                                                  destination,
                                                  partitionNumber,
                                                  receiverListPtr);
    _addOneDefaultListener<runcontrolmessages::BeginRunRequest,
        runcontrolmessages::BeginRunResponse>(ddsConnection,
                                              destination,
                                              partitionNumber,
                                              receiverListPtr);
    _addOneDefaultListener<runcontrolmessages::PauseRunRequest,
        runcontrolmessages::PauseRunResponse>(ddsConnection,
                                              destination,
                                              partitionNumber,
                                              receiverListPtr);
    _addOneDefaultListener<runcontrolmessages::ResumeRunRequest,
        runcontrolmessages::ResumeRunResponse>(ddsConnection,
                                               destination,
                                               partitionNumber,
                                               receiverListPtr);
    _addOneDefaultListener<runcontrolmessages::StopRunRequest,
        runcontrolmessages::StopRunResponse>(ddsConnection,
                                             destination,
                                             partitionNumber,
                                             receiverListPtr);
    _addOneDefaultListener<runcontrolmessages::SetParticipantsRequest,
        runcontrolmessages::SetParticipantsResponse>(ddsConnection,
                                             destination,
                                             partitionNumber,
                                             receiverListPtr);
}

};  // end of namespace rcclient

};  // end of namespace novadaq
