/*
 * SNEWSMessageForwarder
 * Receive SNEWS message information and distribute to recipients as specified in config.
 *  
 * Original:  
 * @author Holger Meyer
 * @version $Revision: 1.2 $ $Date: 2014/09/10 15:12:16 $
 *
 * Adapted for SNEWS:
 * J. Zirnstein (janzirn@fnal.gov)
 */
#include "XmlRpc/XmlRpc.h"
#include "NovaSNEWSInterface/SNEWSXmlRpcServerClasses.h"
#include "NovaSNEWSInterface/SNEWSMessageInfo.h"
//#include "NovaSpillServer/NssUtil.h"
#include "NovaSNEWSInterface/SNEWSMessageForwarderConfig.h"
extern "C"{
#include "NovaSNEWSInterface/daemon_init.h"
}
#include <ctime>
#include <sstream>
#include <iostream>
#include <fstream>
#include <string>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
// Include the message facility
#include <messagefacility/MessageLogger/MessageLogger.h>

#include <NovaDAQUtilities/LocationUtils.h>

#include <boost/program_options.hpp>
namespace po = boost::program_options;

using namespace nova::NovaSNEWSInterface;
using namespace novadaq::nsi;

#ifndef TRUE
#define TRUE   1
#endif /* TRUE */
	
	
#ifndef FALSE	
#define FALSE  0	
#endif /* FALSE */

int main(int argc, char* argv[])
{


  // Parse command line options using Boost
  std::string cfg_file, sourceHost, destinationHost;
  int sourcePort = -1;
  int destinationPort = -1;
  int inport = -1;
  int daemon_flag = FALSE;  //flag to run process in daemon mode
  int message_flag = FALSE; //flag to log in the message facility

  // Declare the supported options.
  po::options_description desc("Allowed options");
  desc.add_options()
    ("help,h", "usage descriptions")
    ("config,c", po::value<string>(&cfg_file)->default_value("srt://NovaSNEWSInterface/config/SNEWSMessageForwarderConfig.xml"),
     "The configuration xml file to use")
    ("port,p", po::value<int>(&inport)->default_value(-1),
     "The port number to receive xmlrpc SNEWS messages on, must be specified here or in the config file.")
    ("source-port,s", po::value<int>(&sourcePort)->default_value(-1),
     "The port number of a SNEWS-message-forwarder to request messages from.")
    ("source-hostname,n", po::value<std::string>(&sourceHost)->default_value(""),
     "The hostname of a SNEWS-message-forwarder to request messages from.")
    ("destination-port,d", po::value<int>(&destinationPort)->default_value(-1),
     "The port number of a SNEWS-message-forwarder to request messages from.")
    ("destination-hostname,m", po::value<std::string>(&destinationHost)->default_value(""),
     "The hostname of a SNEWS-message-forwarder to request messages from.")
    ("daemon mode,D", po::value<int>(&daemon_flag)->default_value(0),
     "Flag to run non-interactively in daemon mode. Set to true (1) by default, to disable set 0.")
    ("message facility,M", po::value<int>(&message_flag)->default_value(0),
     "Flag to log output to the message facility.  Set to true (1) by default, to disable set 0.");
    ;

  po::variables_map vm;
  po::store(po::parse_command_line(argc, argv, desc), vm);
  po::notify(vm);    

  if (vm.count("help")) {
    cout << desc << "\n";
    exit(1);
  }

  std::string configname = novadaq::LocationUtils::translateURL(cfg_file);

  //Pull in the XML config file
  
  std::string xmlString;
  {
    std::ifstream inStream(configname.c_str());
    if (! inStream.is_open()) {
      // ::mf::LogError("Configuration | FileNotFound") 
      std::cerr << "Configuration | FileNotFound \n"
		<< "Unable to open " << configname << "." << std::endl;
    }
    else {
      std::string line;
      while (std::getline(inStream, line)) {
	xmlString.append(line);
      }
      inStream.close();
    }
  }
  
  //connect objects in XSD file to settings in XML file
  SNEWSMessageForwarderConfigSet SNEWSMFCfgSet;
  try {
    SNEWSMFCfgSet.deserialize(xmlString);
  }
  //catch (novadaq::cfg::xsd::cxx::tree::parsing<char>& excpt) {
  catch (...) {
    std::cerr << "Configuration | Deserialization failed\n"
              << "Syntax error in your configuration xml file?\n"
              << " File content (" << configname << ") :\n"
              << xmlString << "\n Exception "
//            << excpt 
              << std::endl;
    exit (-1);
  }
  
  int mfCfgCount = SNEWSMFCfgSet.getSNEWSMessageForwarderConfigCount();
  if (mfCfgCount > 1) {
    std::cout << "Expect 1 configuration, but " << mfCfgCount
	      << " are specified in the configuration file.\n"
	      << "Using first configuration only!" << std::endl;
  }
  if (mfCfgCount == 0) {
    std::cout << "Try again with a valid configuration. Bye!" << std::endl;
    exit(-1);
  }

  SNEWSMessageForwarderConfigType const& SNEWSMFCfg =
    SNEWSMFCfgSet.getSNEWSMessageForwarderConfig(0);

  //daemonizing the application so this can safely run in the background
  if (daemon_flag){ daemon_init(); }

  // Set up a file for logging
  std::string logFname;
  {
    std::stringstream fname;
    timeval now;
    struct tm* ptm;
    char timeStr[40];
    gettimeofday(&now,0);
    ptm = localtime (&now.tv_sec);
    strftime (timeStr, sizeof (timeStr), "%Y-%m-%d_%H-%M-%S", ptm);
    fname<<"/daqlogs/NovaSNEWSInterface/SNEWSMessageForwarder/SNEWSMessageForwarder"<<"_"<<timeStr<<"_log.txt";
    logFname=fname.str();
  }

  // Start the Message Facility
  if(message_flag){
    mf::StartMessageFacility(mf::MessageFacilityService::MultiThread,
//			     mf::MessageFacilityService::logFile(logFname));
			     mf::MessageFacilityService::logConsole());
    mf::SetApplicationName("SNEWSMessageForwarder");
    mf::SetModuleName("Main");
  }

  // Get exactly one Port to receive SNEWS information if it was not
  // specified on the command line
  if (inport == -1) inport = SNEWSMFCfg.getSNEWSSource().getPort();

  std::cout << "Forwarding SNEWS messages received on port " << inport << "."
            << std::endl;
  ::mf::LogInfo("sPort")<<"Forwarding SNEWS messages received on port "<<inport;

  if (destinationPort != -1 || destinationHost != "") {
    if (destinationPort > 1024 && destinationPort <= 65535) {
      DestList::Instance().addDestination(destinationHost.c_str(),
					  destinationPort,
					  SNEWSMessageInfo::allMessageTypes());
    } else {
      std::cout << "SNEWSDestination " << destinationHost << ":"
		<< destinationPort << " NOT added. Please specify a valid port."
		<< std::endl;
      ::mf::LogError("validPort")<<"SNEWS message Destination "<<destinationHost<<":"<<destinationPort
                                 <<" NOT added.  Please specify a valid port.";
    }
  }

  // Serve to 1 or more (up to 10) destinations as given in config file.
  SNEWSDestinationSetType const& sdestCfgSet = SNEWSMFCfg.getSNEWSDestinationSet(0);
  int sdestCfgCount = sdestCfgSet.getSNEWSDestinationCount();
  std::cout << "Found "<<sdestCfgCount<<" destinations."<<std::endl;
  ::mf::LogInfo("Initialize") << sdestCfgCount << " destinations." << std::endl;
  for (int i = 0; i < sdestCfgCount; ++i) {
    SNEWSDestinationType const& sdestCfg = sdestCfgSet.getSNEWSDestination(i);
    DestinationType const& destCfg = sdestCfg.getDestination();
    std::string host = destCfg.getHost();
    int port = destCfg.getPort();
    bool enabled = destCfg.getEnable();
    if (!enabled) {
      std::cout << "Destination "<<host<<":"<<port<<" is disabled."<<std::endl;
      ::mf::LogInfo("Initialize") << "Destination "<<host<<":"<<port<<" is disabled.";
    } else {
      std::cout << "Destination "<<host<<":"<<port<<" will receive SNEWS messages ";
      ::mf::LogInfo("Initialize") << "Destination "<<host<<":"<<port<<" will receive SNEWS messages ";
      SNEWSTypeSetType const& stst =  sdestCfg.getSNEWSTypeSet();
      std::list<int> messagetypes;
      for (int j=0; j!= stst.getSNEWSTypeCount(); ++j) {
        SNEWSTypeType st = stst.getSNEWSType(j);
	std::cout << st.getName();
	if (j<stst.getSNEWSTypeCount()-1) {std::cout << ", ";}
       	int type = SNEWSMessageInfo::getMessageTypeFromName(std::string(st.getName()));
        if (find(messagetypes.begin(),messagetypes.end(),type) == messagetypes.end()
	    && type>-1) {
	  messagetypes.push_back(type);
	}
      }
      std::cout <<std::endl;

      DestList::Instance().addDestination(host.c_str(),port,messagetypes);

    }
  }

  XmlRpc::setVerbosity(0);
  XmlRpc::XmlRpcServer s;

  SNEWSMessageForward messageForward(&s);
  DestinationAdd destinationAdd(&s);
  DestinationRemove destinationRemove(&s);

  s.bindAndListen(inport);

  // Send request to receive SNEWS messages 
  if (sourcePort>1024 && sourcePort <=65535 && sourcePort != inport &&
      sourceHost != "") {
    // See if a forwarder is listening on sourcehost:sourceport.
    // If so, ask for SNEWS messages to be sent to localhost:inport, i.e. this process
    XmlRpc::setVerbosity(0);
    std::string thisHostname;

    //Get the hostname we are on
    struct addrinfo hints, *info, *p;
    int gai_result;
    
    char hostname[1024];
    hostname[1023] = '\0';
    gethostname(hostname, 1023);
    
    memset(&hints, 0, sizeof hints);
    hints.ai_family = AF_UNSPEC; /*either IPV4 or IPV6*/
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags = AI_CANONNAME;
    
    if ((gai_result = getaddrinfo(hostname, "http", &hints, &info)) != 0) {
      fprintf(stderr, "getaddrinfo: %s\n", gai_strerror(gai_result));
      exit(1);
    }
    
    for(p = info; p != NULL; p = p->ai_next) {
      printf("local hostname: %s\n", p->ai_canonname);
      //How to select the right one if we are on more than one network?
      thisHostname = p->ai_canonname;
    }
    
    freeaddrinfo(info);

    //Now send the request to receive SNEWS messages
    XmlRpc::XmlRpcClient c(sourceHost.c_str(),sourcePort);
    XmlRpc::XmlRpcValue result, params;
    params[0] = thisHostname;
    params[1] = inport;
    std::cout << "Sending DestinationAdd(" << thisHostname << ":"
	      << inport << ")..." << std::endl;
    ::mf::LogInfo("Initialize") << "Sending DestinationAdd(" << thisHostname << ":"
	                  << inport << ")...";
    c.execute("DestinationAdd",params,result);
    if (std::string(result).size()!=0) {
      std::cout << "DestinationAdd result: " << result <<std::endl;
      ::mf::LogInfo("Initialize") << "DestinationAdd result: " << (std::string)result;
    } else {
      std::cout << "Registering with a SNEWSMessageSender failed. I will "
		<< "continue to listen for messages\n" << "on port " << inport 
		<< ". You have to make sure someone is going to send them."
		<< std::endl;
      ::mf::LogError("RegisterFail")  << "Registering with a SNEWSMessageForwarder failed. I will "
		              << "continue to listen for messages on port " << inport 
		              << ". You have to make sure someone is going to send them.";
    }
  } else {
    if (sourcePort != -1) {
      std::cout << "Bad parameters to request SNEWS messages. DestinationAdd was skipped." << std::endl;
      ::mf::LogError("BadPort") << "Bad parameters to request SNEWS messages. DestinationAdd was skipped.";
    }
  }

  s.work(-1.0);

  return 0;
}
