/*
 * SNEWSXmlRpcServerClasses implementation file
 */
#include "NovaSNEWSInterface/SNEWSXmlRpcServerClasses.h"
#include "NovaSNEWSInterface/SNEWSMessageInfo.h"
#include "NovaSNEWSInterface/SNEWSUtil.h"
#include "rms/provider/DDSConnection.h"
#include "rms/base/RmsDestination.h"
#include "rms/RmsReceiver.h"
#include "rms/RmsSender.h"
#include "XmlRpc/XmlRpc.h"

#include <boost/thread.hpp>
#include <iostream>
#include <stdlib.h>
#include <string>
#include <utility>

#include <NovaTimingUtilities/TimingUtilities.h>
// Include the message facility
#include <messagefacility/MessageLogger/MessageLogger.h>

#define TRIALLIMIT 3


using namespace XmlRpc;
using namespace gov::fnal::cd::rms;
using namespace nova::NovaSNEWSInterface;

SNEWSMessage::SNEWSMessage (XmlRpcServer* s,
	      boost::shared_ptr<gov::fnal::cd::rms::RmsSender<gov::fnal::cd::rms::provider::DDSConnection,nsimessages::SNEWSMessage> > p) :
  XmlRpcServerMethod("SNEWSMessage", s),
  requestSenderptr(p)
{
}

// Receive SNEWSMessage information and distribute through rms message
void SNEWSMessage::execute(XmlRpcValue& params, XmlRpcValue& result)
{
  std::string const bad("bad");
  int nArgs = params.size();
  if (nArgs != 2) {
    //We expect to receive the time and type here. Something wrong...
    //ToDo: don't just quietly fail...
    result = bad;

    //std::cout << "NovaSNEWSInterface: This is bad:" << std::endl;
    //std::cout << " got "<<nArgs<<" parameters, expected 2." << std::endl;
    for (int i = 0; i != nArgs; ++i) {
      std::cout << "Param" << i << ": " << params[i] << std::endl;
    }
    ::mf::LogError("receiverParams")<<"Expected to receiver two parameters, instead got: "<<nArgs;

    return;
  }

  //Should do more type checking here. The params could be anything...
  SNEWSMessageInfo messageInfo;
  uint64_t ttime;
  SNEWSUtil::gpstimeFromString(ttime,string(params[0]));
  int ttype = int(params[1]);

  messageInfo.set(ttime,ttype);
  nsimessages::SNEWSMessage SNUMessage;
  SNUMessage.time = messageInfo.getTime();
  SNUMessage.type = messageInfo.getType();
  requestSenderptr->sendMessage(SNUMessage);

  //compute difference from unix system time
  timeval now;
  uint64_t unixNow;
  gettimeofday(&now,0);
  novadaq::timeutils::convertUnixTimeToNovaTime(now,unixNow);


  //make the message receiver noisy so we know it is working
  //std::cout << "SNEWSMessage message received of type: "<< messageInfo.getType() << " at time: "
    //        << messageInfo.getTime() << std::endl;
  ::mf::LogInfo("messageReceiver")<< "Received type: "<<messageInfo.getType()<<" triggerTime: "
                                <<messageInfo.getTime()<<" currentTime: "<<unixNow;
  //Don't wait for reply. 
  result = std::string("Ok");
}

Dest::Dest(std::string const& h, int const& p, std::list<int> const& tlist) :
  host(h),
  port(p),
  typelist(tlist),
  client(XmlRpcClient(host.c_str(), port))
{
}

DestList* DestList::fInstance = 0;

DestList& DestList::Instance()
{
  if (!fInstance) {
    fInstance = new DestList;
  }
  return *fInstance;
}

DestList::DestList() {;}

void DestList::addList(std::vector<std::pair<std::basic_string<char>, int> > const& rL)
{
  std::string host;
  int port;

  for (unsigned int i=0; i!= rL.size(); ++i) {
    host = rL[i].first;
    port = rL[i].second;
    recList.push_back(Dest(host.c_str(), port, SNEWSMessageInfo::allMessageTypes()));
  }
}

int DestList::addDestination(std::string const & h,
		     int const & p,
		     std::list<int> const& tlist)
{
  //Check if this destination is already in the list.
  //Note: the duplicate is not detected it the hostname string differs,
  // e.g. with or without full domain
  for (std::vector<Dest>::const_iterator it = recList.begin();
       it != recList.end(); ++it) {
    if (it->host == h && it->port == p) {
      std::cout << "Host " << h << ":" << p 
		<< " was added before. (Remove first to change types of message to forward.)"
		<< std::endl; 
      ::mf::LogWarning("DestAdd")<<"Host "<<h<<":"<<p<<" was added before. (Remove first to change types of message to forward.)";
      return 1;
    }
  }

  //Check if all the requested messages can be forwarded.
  //(Might ask a forwarder to forward message types it itself does not receive.)
  //TODO!!! check list against availableMessageTypes

  recList.push_back(Dest(h.c_str(), p, tlist));

  return 0;
}

void DestList::addAvailableMessageType(int const& t)
{
  //Check range
  if ( t<0 || t>= kNMessageType ) return;
  //Don't add same type twice
  if ( find(availableMessageTypes.begin(),availableMessageTypes.end(),t) == 
       availableMessageTypes.end() ) {
    availableMessageTypes.push_back(t);
  }
}


// SNEWSMessageForward: receive info and send out through xmlrpc again
//
SNEWSMessageForward::SNEWSMessageForward(XmlRpcServer* s) :
  XmlRpcServerMethod("SNEWSMessage", s)
{
}

void SNEWSMessageForward::execute(XmlRpcValue& params, XmlRpcValue& result)
{

  int nArgs = params.size();
  if (nArgs != 2) {
    //We expect to receive the time and type here. Something wrong...
    //ToDo: don't just quietly fail...
    result = std::string("Did not receive exactly 2 parameters. SNEWSMessageForward failed.");
    ::mf::LogWarning("forwardArgs")<<"Did not receive exactly 2 paramters.  SNEWSMessageForward failed.";
    return;
  }

  //Should do more type checking here. The params could be anything...

  std::vector< Dest > dList = DestList::Instance().ReceiverList();

  //XmlRpcValue res;

  //Forward to all receivers
  for (unsigned int i=0; i!=dList.size(); ++i) {
    //Check if this type of message should be forwarded to this destination
    Dest const& d = dList[i];
    if ( find(d.typelist.begin(),d.typelist.end(),(int)params[1]) != 
	 d.typelist.end() ){

      // Execute call to one of the forwarder destinations
      // this call will normally block until something happens
      // We don't like this so we wrapper it in a thread
      boost::thread DispatchSNEWSMessageThread(&SNEWSMessageForward::DispatchTheSNEWSMessage, this, params, i);

    }
  }

  //Don't wait for reply. 
  result = std::string("Ok");
}

void SNEWSMessageForward::DispatchTheSNEWSMessage(XmlRpcValue& params, int i ){
  XmlRpcValue res;
  std::vector< Dest > dList = DestList::Instance().ReceiverList();

  //compute difference from unix system time
  timeval now;
  uint64_t unixNow, unixNow2;
  gettimeofday(&now,0);
  novadaq::timeutils::convertUnixTimeToNovaTime(now,unixNow);

  uint64_t ttime;
  SNEWSUtil::gpstimeFromString(ttime,string(params[0]));
  int ttype = int(params[1]);

  //if the message fails to be sent, try again before proceeding
  int trialCount = 0;
  bool rval = false;
  while (!rval){
    rval = dList[i].client.execute("SNEWSMessage",params, res);
    trialCount++;
    
    //trial limit defined as 2 for now
    if (trialCount == TRIALLIMIT) break;
  }
    
    //if (dList[i].client.
  std::cout << "SNEWSMessage forwarded: "<< params[0] << " "
	    << params[1] << " " << rval << " " << trialCount <<std::endl;
  gettimeofday(&now,0);
  novadaq::timeutils::convertUnixTimeToNovaTime(now,unixNow2);
  ::mf::LogInfo("messageForward")<<"SNEWSMessageType: "<<ttype<<" triggerTime: "<<ttime
                               <<" forwardTimeBegin: "<<unixNow<<" forwardTimeEnd: "<<unixNow2
                               <<" destination: "<<i<<" rval: "<<rval<<" tcount: "<<trialCount;

  //closing the connection after message is sent
  //must be done otherwise open connections will pile up
  //and cause errors when the maximum number of open files is reached
  dList[i].client.close();
}


DestinationAdd::DestinationAdd(XmlRpcServer* s) :
  XmlRpcServerMethod("DestinationAdd", s)
{
}

void DestinationAdd::execute(XmlRpcValue& params, XmlRpcValue& result)
{
  bool printdebug = true;

  //params should contain host, port, and optionally list of messagetypes

  int nArgs = params.size();
  if (nArgs < 2) {
    result = std::string("Received less than 2 parameters. DestinationAdd failed.");
    return;
  }

  //Should do more type checking here. The params could be anything...
  std::string host = params[0];
  int port = params[1];
  if ( port<=1024 || port>65535 ) {
    result = std::string("Received bad port number. DestinationAdd failed.");
    return;
  }
  
  std::list<int> tList;
  if (nArgs == 2) {
    //tList = DestList::Instance().AvailableMessageTypes();
    //implement the above? For now just forward all
    tList = SNEWSMessageInfo::allMessageTypes();

  }
  else {
    for (int i=2; i!=nArgs; ++i) {
      tList.push_back((int)params[i]);
    }
  }

  if (printdebug)
    std::cout << "DestinationAdd: " << host << ":" << port << std::endl;

  int r = DestList::Instance().addDestination(host, port, tList);
  if (r==0) {
    result = std::string("Ok");
  }
  else {
    result = std::string("not all is well");
  }
}

DestinationRemove::DestinationRemove(XmlRpcServer* s) :
  XmlRpcServerMethod("DestinationRemove", s)
{
}

void DestinationRemove::execute(XmlRpcValue& params, XmlRpcValue& result)
{
  //TODO
}
