/**
 * NssTDUApp
 * $Id: NssStandAloneDecoder.cc,v 1.1 2013/02/19 19:10:19 anorman Exp $
 * The application that has to run on the Master TDU to serve spills
 * 
 * @author Holger Meyer
 * @version $Revision: 1.1 $ $Date: 2013/02/19 19:10:19 $
 */
#include "NovaSpillServer/NssSpillInfo.h"
#include "NovaSpillServer/NssUtil.h"
#include "NovaSpillServer/NssTDUAppConf.h"
#include "NovaSpillServer/NssTDUAppConfig.h"
#include "XmlRpc/XmlRpc.h"
#include "NovaTimingUtilities/TimingUtilities.h"
#include "NovaDAQUtilities/LocationUtils.h"
#include "TDUUtilities/tdu.h"
#include <iostream>
#include <string>
#include <fstream>
#include <ctime>
#include <list>
#include <boost/program_options.hpp>
namespace po = boost::program_options;

using namespace nova::NovaSpillServer;
using namespace novadaq::nss;

static uint64_t const fifteensecs = 15*64000000;

int main(int argc, char* argv[])
{
  // Parse command line options using Boost
  std::string cfg_file;

  // Declare the supported options.
  po::options_description desc("Allowed options");
  desc.add_options()
    ("help,h", "usage descriptions")
    ("config,c",
     po::value<std::string>(&cfg_file)->
        default_value("srt://NovaSpillServer/config/NssTDUAppConfig.xml"),
     "The configuration xml file to use")
    ;

  po::variables_map vm;
  po::store(po::parse_command_line(argc, argv, desc), vm);
  po::notify(vm);    

  if (vm.count("help")) {
    std::cout << desc << "\n";
    exit(1);
  }

  std::string fname = novadaq::LocationUtils::translateURL(cfg_file);

  //Pull in the XML config file
  
  std::string xmlString;
  {
    std::ifstream inStream(fname.c_str());
    if (! inStream.is_open()) {
      // ::mf::LogError("Configuration | FileNotFound") 
      std::cerr << "Configuration | FileNotFound \n"
		<< "Unable to open " << fname << "." << std::endl;
      exit(-1);
    }
    else {
      std::string line;
      while (std::getline(inStream, line)) {
	xmlString.append(line);
      }
      inStream.close();
    }
  }
 
  //connect objects in XSD file to settings in XML file
  novadaq::nss::NssTDUAppConfigType NssTACfg;
  try {
    NssTACfg.deserialize(xmlString);
  }
  //catch (novadaq::cfg::xsd::cxx::tree::parsing<char>& excpt) {
  catch (...) {
    std::cerr << "Configuration | Deserialization failed\n"
              << "Syntax error in your configuration xml file?\n"
              << " File content (" << fname << ") :\n"
              << xmlString << "\n Exception "
//            << excpt 
              << std::endl;
    exit (-1);
  }
  
  novadaq::nss::TDUSpillDestinationType const& destCfg = 
    NssTACfg.getTDUSpillDestination();

  std::string rpcHost = destCfg.getTDUDestination().getHost();
  int rpcPort = destCfg.getTDUDestination().getPort();

  std::cout << "Sending spill information to " << rpcHost << ":"
            << rpcPort << std::endl;

  TDUSpillTypeSetType const& tsts = destCfg.getTDUSpillTypeSet();

  //get list of spilltypes from configuration
  std::list<int> sptypes;
  for (int j=0; j!= tsts.getTDUSpillTypeCount(); ++j) {
    TDUSpillTypeType st = tsts.getTDUSpillType(j);
    int type = NssSpillInfo::getSpillTypeFromName(std::string(st.getName()));
    if (find(sptypes.begin(),sptypes.end(),type) == sptypes.end()
        && type>-1) {
      sptypes.push_back(type);
    }
  }

  //Now take more detailed configuration options into account
  //and put the event mask together
  novadaq::nss::TDUAppConfType const& tappCfg = NssTACfg.getTDUAppConf();

  //see description of TDU registers in nova-doc-5370.
  //Should get these from some header file:
  uint32_t evtMask = 0x0; //set below, write to reg. 0x23
  uint32_t ctrlMask = 0x0400; //bit 10: enable events, write to reg. 0x0
  uint32_t irqMask = 0x0016; //bit 4: enable PowerPC interrupt, to reg 0x24

  uint32_t enableA9 = (1<<0);
  uint32_t enableAD = (1<<1);
  uint32_t enable1F = (1<<2);
  uint32_t enable1D = (1<<3);
  uint32_t enableNuMI = (1<<8);
  uint32_t enableBNB = (1<<12);

  //NuMI and BNB
  if (find(sptypes.begin(),sptypes.end(),kNuMI) != sptypes.end()) {
    evtMask |= enableNuMI;
  }
  if (find(sptypes.begin(),sptypes.end(),kBNB) != sptypes.end()) {
    evtMask |= enableBNB;
  }
  //NuMItclk
  if (find(sptypes.begin(),sptypes.end(),kNuMItclk) != sptypes.end()) {
    if (tappCfg.NuMItclkIssueOnA9()) {
      evtMask |= enableA9;
    }
    if (tappCfg.NuMItclkIssueOnAD()) {
      evtMask |= enableAD;
    }
    if ((!tappCfg.NuMItclkIssueOnA9()) && (!tappCfg.NuMItclkIssueOnAD())) {
      std::cout << "Warning: configuration asks for NuMItclk spills,"
                << " but has both $AD and $A9 disabled. Choose one!"
                << std::endl;
    }
  }
  //BNBtclk
  if (find(sptypes.begin(),sptypes.end(),kBNBtclk) != sptypes.end()) {
    evtMask |= enable1F;
    if (tappCfg.BNBtclkuse1Dand1F()) evtMask |= enable1D;
  }

  std::cout << "Spills to generate: ";
  for (std::list<int>::const_iterator it = sptypes.begin();
       it != sptypes.end();   ++it)
  {
    std::cout << NssSpillInfo::getSpillNameFromType((*it)) << " ";
  }
  std::cout << std::endl;

  // get some configuration
  NssTDUAppConf cnf;

  // set up XmlRpcClient
  XmlRpc::setVerbosity(0);
  XmlRpc::XmlRpcClient c(rpcHost.c_str(),
                         rpcPort);
  XmlRpc::XmlRpcValue result;

  // set up the TDU to send spills based on configuration
  nova::TDUUtilities::tdu mtu;

  // For a while the TDU firmware converted GPS time into nova-time using
  // a nova-epoch that was wrong by 15 seconds. (# of leap seconds)
  // This is fixed with the newer firmware versions.
  uint32_t FPGAFirmwareVersion = mtu.readFPGAFirmwareVersion();
  //std::cout << "Firmware version 0x" << std::hex << FPGAFirmwareVersion
  //          << std::dec << " detected." << std::endl;
  bool subfifteen = (FPGAFirmwareVersion <= 0x0106);

  uint32_t ctrl = mtu.getRegister(cnf.ctrlReg);
  ctrl |= ctrlMask;
  mtu.setRegister(cnf.ctrlReg,ctrl);
 
  mtu.setRegister(cnf.evtReg,evtMask);

  uint32_t irq = mtu.getRegister(cnf.irqReg);
  irq |= irqMask;
  mtu.setRegister(cnf.irqReg,irq);

  std::string str;
  SpillType spillType;
  XmlRpc::XmlRpcValue spillInfoXml;

  // start to send out spills
  uint16_t buffer[4096];
  
  std::cout << "Sending spills..." << std::endl;

  while (true) {

    int bytesRead = mtu.readEvents((uint8_t *)buffer, 4096);

    //printf ("bytes read = %d\n", bytesRead);
    if (bytesRead%10 != 0) {
      std::cout << "I expected to read an integer multiple of 10 bytes.\n"
                << "I got " << bytesRead << " bytes instead.\n"
                << "This is weird. Alert an expert!!"<<std::endl;
    }

    uint64_t latest1D = 0x0; //keep track of 1Ds for BNBtclk

    for (int i=0; i<bytesRead/10; ++i)
    {
      bool send = true;
      uint16_t evt = buffer[i*5+0];
      uint64_t novatime = 0;
      for (int j=0; j!=4; ++j) {
        novatime += ((uint64_t)buffer[i*5+(j+1)])<<(16*j);
      }

      //Subtract 15 seconds if we run with old firmware
      if (subfifteen) novatime -= fifteensecs;

      //decode time into spillInfoXml
      NssUtil::gpstimeToString(novatime, str);
      spillInfoXml[0] = str;

      //decode evt into spillInfoXml
      NssSpillInfo::getSpillTypeFromEvent(spillType,evt);
      spillInfoXml[1] = spillType;

      if (tappCfg.BNBtclkuse1Dand1F()) {
        if ((evt & 0xFF) == 0x1D) {
          //Don't issue $1D. Just update to form delayed coincidence
          latest1D = novatime;
          std::cout << "Not sent: 0x" << std::hex << evt << " at 0x"
                    << std::hex << novatime << std::dec << std::endl;
          send = false;
        } else if ((evt & 0xFF) == 0x1F) {
          if ((novatime < (latest1D + tappCfg.BNBtclk1Ddelay())) ||
              (novatime > (latest1D + tappCfg.BNBtclk1Ddelay()
	                            + tappCfg.BNBtclk1Dwidth())) )
          {
            //This $19 does not fall into the window after a $1D. Skip.
	    std::cout << "Not sent: 0x" << std::hex << evt << " at 0x"
                      << std::hex << novatime << std::dec << std::endl;
            send = false;
          }
        }
      }

      // Instead of sending the spills on to a destination, we 
      // just decode them and print the result to the console
      // This allows us the lowest level access to the spill times
      
      if(send){
	printf("0x%04x,0x%16Lx,0x%16Lx\n",evt,novatime,novatime); 
      }

#if 0
      if (send) {
        if (c.execute("Spill", spillInfoXml, result)) {
          //std::cout << "Spill method result: " << result << "\n";
          std::cout << "Type " 
                    << NssSpillInfo::getSpillNameFromType(spillType)
                    << "(evt 0x" << std::hex << evt << ") at 0x"
                    << std::hex << novatime << std::dec << std::endl;
        }
        else {
          std::cout << "NssTDUApp: Error sending 'Spill'\n";
        }
      }
#endif
    } // endfor
    sleep(1);
  }
 
  return 0;
}
