/**
 * NssTDUApp
 * $Id: TCRMonitor.cc,v 1.8 2015/05/22 21:01:47 anorman Exp $
 * The application that has to run on the Master TDU to serve spills
 * 
 * @author Holger Meyer
 * @version $Revision: 1.8 $ $Date: 2015/05/22 21:01:47 $
 */
// System C Libs
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <unistd.h>
// System C++ Libs
#include <iostream>
#include <string>
#include <fstream>
#include <ctime>
#include <list>
// External Libs
#include <boost/program_options.hpp>
// Custom Libs
#include "XmlRpc/XmlRpc.h"
#include "TDUUtilities/tdu.h"
#include "NovaTimingUtilities/TimingUtilities.h"
#include "NovaDAQUtilities/LocationUtils.h"
#include "NovaSpillServer/NssSpillInfo.h"
#include "NovaSpillServer/NssUtil.h"
#include "NovaSpillServer/NssTDUAppConf.h"
#include "NovaSpillServer/NssTDUAppConfig.h"
extern "C"{
#include "NovaSpillServer/daemon_init.h"
}
// Include the message facility
#include <messagefacility/MessageLogger/MessageLogger.h>

// Manage the Namespaces
namespace po = boost::program_options;

using namespace nova::NovaSpillServer;
using namespace novadaq::nss;

// Define Macros for managing the Shared Memory Segment
#define SHM_MODE (0666)
#define SHM_MODE_CREAT (IPC_CREAT | 0666)

#define SHM_SIZE 0x80000  /* 512k */
#define DEFAULT_MEM_SIZE  0x2000000

#ifndef TRUE
#define TRUE   1
#endif /* TRUE */

#ifndef FALSE
#define FALSE  0
#endif /* FALSE */

#ifndef FAIL
#define FAIL  -1
#endif /* FAIL */

// Global Constants
static uint64_t const fifteensecs = 15*64000000;

// Global Variables
uint64_t lastspill_time;
uint64_t lastTCR_time;
uint64_t tcrCount;

struct AccelEventData {
  uint32_t evt_type;
  uint32_t evt_no;
  uint64_t time;
};

struct SHMSpillDataMap {
  uint32_t* data_start;
  uint32_t* data_end;
  struct AccelEventData* data_insert;
  uint32_t data_entries;
};

int main(int argc, char* argv[])
{
  // Initialize Variables
  lastspill_time = 0;
  lastTCR_time = 0;
  tcrCount = 0;

  // int    memseg_flag    = FALSE;                         // Flag indicating we set a mem segement id
  int    memid          = -1;                            // Base ID for the shm segment (0x0 = IPC_PRIVATE)
  // unsigned int basesize = 0x0000400;                     // Base size for the initial shm query (1k)

  unsigned int memsize  = DEFAULT_MEM_SIZE;              // Default size of shared memory segment is 32M
  int    shmid          = 0;                             // Shared memory ID
  // struct shmid_ds         shm_stats;                     // Data structure for SHM segment stats
  char*  shmptr         = NULL;                          // Pointer to the start of the shared memory buffer
  // int*   shmptr_ins     = NULL;                       // Offset for insertion into the SHM buffer
  int*   shmptr_int     = NULL;                          // Pointer (integer word) to the start of the shared memory buffer
  char*  shmptr_start   = NULL;                          // Pointer to the start of the data area in the shared memory buffer
  char*  shmptr_end     = NULL;                          // Pointer to the end of the data area in the shared memory buffer
  char*  shmptr_current = NULL;                          // Pointer to the current insertion point in the data buffer

  struct SHMSpillDataMap* DataMap;                       // Pointer to the Shared Memory Data Map
  struct AccelEventData*  AccelData;                     // Pointer to the start of the Accelerator Data block

  AccelEventData theAccelData;                           // Structure to hold the current accelerator event
  theAccelData.evt_type = 0;
  theAccelData.time     = 0;
  theAccelData.evt_no   = 0;

  int daemon_flag = TRUE; //flag to run process in daemon mode
  int message_flag = TRUE; //flag to log output in the message facility
  int server_flag = TRUE; //flag to log output to server in addition to console

  // Parse command line options using Boost
  std::string cfg_file;

  // Declare the supported options.
  po::options_description desc("Allowed options");
  desc.add_options()
    ("help,h", "usage descriptions")
    ("config,c",
     po::value<std::string>(&cfg_file)->
        default_value("srt://NovaSpillServer/config/NssTDUAppConfig.xml"),
     "The configuration xml file to use")
    ("message facility,M", po::value<int>(&message_flag),
     "Flag to log output in the message facility. Set to true (1) by default, to disable set 0.")
    ("daemon mode,D", po::value<int>(&daemon_flag),
     "Flag to run non-interactively in daemon mode. Set to true (1) by default, to disable set 0.")
    ("server mode,S", po::value<int>(&server_flag),
     "Flag to log messages to server (requires running DDS) in addition to console.  Set to true (1) be default, to disable set 0.")
#if 0
    ("shm,s",po::value<int>(&memid)->default_value(0x0),
     "Memory Segment identifier to connect to (hex)"),
    ("mem,m",po::value<int>(&memid)->default_value(0x0),
     "Memory Segment identifier to connect to (4 char)")
#endif
    ;

  po::variables_map vm;
  po::store(po::parse_command_line(argc, argv, desc), vm);
  po::notify(vm);    

  if (vm.count("help")) {
    std::cout << desc << "\n";
    exit(1);
  }

  //daemonizing the application so this can safely run in the background
  if (daemon_flag){
    daemon_init();
  }

  // Retrieve the hostname of the TDU we are running on
  char hostname[1024];
  size_t len=1024;
  gethostname(hostname, len);


  std::string fname = novadaq::LocationUtils::translateURL(cfg_file);

  //Pull in the XML config file
  
  std::string xmlString;
  {
    std::ifstream inStream(fname.c_str());
    if (! inStream.is_open()) {
      // ::mf::LogError("Configuration | FileNotFound") 
      std::cerr << "Configuration | FileNotFound \n"
		<< "Unable to open " << fname << "." << std::endl;
      exit(-1);
    }
    else {
      std::string line;
      while (std::getline(inStream, line)) {
	xmlString.append(line);
      }
      inStream.close();
    }
  }
 
  //connect objects in XSD file to settings in XML file
  novadaq::nss::NssTDUAppConfigType NssTACfg;
  try {
    NssTACfg.deserialize(xmlString);
  }
  //catch (novadaq::cfg::xsd::cxx::tree::parsing<char>& excpt) {
  catch (...) {
    std::cerr << "Configuration | Deserialization failed\n"
              << "Syntax error in your configuration xml file?\n"
              << " File content (" << fname << ") :\n"
              << xmlString << "\n Exception "
//            << excpt 
              << std::endl;
    exit (-1);
  }

  // Start the Message Facility
  if(message_flag){
    if (server_flag){
      mf::StartMessageFacility(mf::MessageFacilityService::MultiThread,
			       mf::MessageFacilityService::logCS());
    }
    else{
      mf::StartMessageFacility(mf::MessageFacilityService::MultiThread,
			       mf::MessageFacilityService::logConsole());
    }
    mf::SetApplicationName("NssTDUApp");
    mf::SetModuleName("Main");
  }
  
  novadaq::nss::TDUSpillDestinationType const& destCfg = 
    NssTACfg.getTDUSpillDestination();

  std::string rpcHost = destCfg.getTDUDestination().getHost();
  int rpcPort = destCfg.getTDUDestination().getPort();

  TDUSpillTypeSetType const& tsts = destCfg.getTDUSpillTypeSet();

  //get list of spilltypes from configuration
  std::list<int> sptypes;
  for (int j=0; j!= tsts.getTDUSpillTypeCount(); ++j) {
    TDUSpillTypeType st = tsts.getTDUSpillType(j);
    int type = NssSpillInfo::getSpillTypeFromName(std::string(st.getName()));
    if (find(sptypes.begin(),sptypes.end(),type) == sptypes.end()
        && type>-1) {
      sptypes.push_back(type);
    }
  }

  //Now take more detailed configuration options into account
  //and put the event mask together
  novadaq::nss::TDUAppConfType const& tappCfg = NssTACfg.getTDUAppConf();

  //see description of TDU registers in nova-doc-5370.
  //Should get these from some header file:
  uint32_t evtMask = 0x0; //set below, write to reg. 0x23
  uint32_t ctrlMask = 0x0400; //bit 10: enable events, write to reg. 0x0
  uint32_t irqMask = 0x0016; //bit 4: enable PowerPC interrupt, to reg 0x24

  uint32_t enableA9 = (1<<0); //TCLK reflection of $74
  uint32_t enableAD = (1<<1); //NUMI reset for beamline ramp and extracted beam
  uint32_t enable1F = (1<<2); //Booster extraction sync for a beam cycle
  uint32_t enable1D = (1<<3); //Booster reset for Miniboone beam cycle
  uint32_t enable8f = (1<<4); //1Hz clock generated by GPS
  uint32_t enable00 = (1<<5); //Super cycle and master clock reset
  uint32_t enableA4 = (1<<6); //NUMI cycle sample trigger, reference for $A5
  uint32_t enableA5 = (1<<7);  //NUMI reset for beam
  uint32_t enableNuMI = (1<<8); //MIBS $74, proton extration into NUMI
  uint32_t enableBNB = (1<<12); //$1B parasitic beam inhibit

  // 1Hz Accel Pulser
  evtMask |= enable8f;

  //NuMI and BNB
  if (find(sptypes.begin(),sptypes.end(),kNuMI) != sptypes.end()) {
    evtMask |= enableNuMI;
  }
  if (find(sptypes.begin(),sptypes.end(),kBNB) != sptypes.end()) {
    evtMask |= enableBNB;
  }
  //NuMItclk
  if (find(sptypes.begin(),sptypes.end(),kNuMItclk) != sptypes.end()) {
    if (tappCfg.NuMItclkIssueOnA9()) {
      evtMask |= enableA9;
    }
    if (tappCfg.NuMItclkIssueOnAD()) {
      evtMask |= enableAD;
    }
    if ((!tappCfg.NuMItclkIssueOnA9()) && (!tappCfg.NuMItclkIssueOnAD())) {
      std::cout << "Warning: configuration asks for NuMItclk spills,"
                << " but has both $AD and $A9 disabled. Choose one!"
                << std::endl;
    }
  }
  //BNBtclk
  if (find(sptypes.begin(),sptypes.end(),kBNBtclk) != sptypes.end()) {
    evtMask |= enable1F;
    if (tappCfg.BNBtclkuse1Dand1F()) evtMask |= enable1D;
  }
  //SuperCycle
  if (find(sptypes.begin(),sptypes.end(),kSuperCycle) != sptypes.end()) {
    evtMask |= enable00;
  }
  //NuMISampleTrig
  if (find(sptypes.begin(),sptypes.end(),kNuMISampleTrig) != sptypes.end()) {
    evtMask |= enableA4;
  }
  //NuMIReset
  if (find(sptypes.begin(),sptypes.end(),kNuMIReset) != sptypes.end()) {
    evtMask |= enableA5;
  }

  std::cout << "Spills to generate: ";
  for (std::list<int>::const_iterator it = sptypes.begin();
       it != sptypes.end();   ++it)
  {
    std::cout << NssSpillInfo::getSpillNameFromType((*it)) << " ";
  }
  std::cout << std::endl;

  // get some configuration
  NssTDUAppConf cnf;

  // set up XmlRpcClient
  XmlRpc::setVerbosity(0);
  XmlRpc::XmlRpcClient c(rpcHost.c_str(),
                         rpcPort);
  XmlRpc::XmlRpcValue result;

  // set up the TDU to send spills based on configuration
  nova::TDUUtilities::tdu mtu;

  // For a while the TDU firmware converted GPS time into nova-time using
  // a nova-epoch that was wrong by 15 seconds. (# of leap seconds)
  // This is fixed with the newer firmware versions.
  uint32_t FPGAFirmwareVersion = mtu.readFPGAFirmwareVersion();
  //std::cout << "Firmware version 0x" << std::hex << FPGAFirmwareVersion
  //          << std::dec << " detected." << std::endl;
  bool subfifteen = (FPGAFirmwareVersion <= 0x0106);

  uint32_t ctrl = mtu.getRegister(cnf.ctrlReg);
  ctrl |= ctrlMask;
  mtu.setRegister(cnf.ctrlReg,ctrl);
 
  mtu.setRegister(cnf.evtReg,evtMask);

  uint32_t irq = mtu.getRegister(cnf.irqReg);
  irq |= irqMask;
  mtu.setRegister(cnf.irqReg,irq);

  std::string str;
  SpillType spillType;
  XmlRpc::XmlRpcValue spillInfoXml;

  // start to send out spills
  uint16_t buffer[4096];


  /****************************************************
   * Setup a shared memory segment to give us a small *
   * history/spy buffer that we can look at to parse  *
   * out beam spill times                             *
   ****************************************************/
  char seg[]="BEAM\0";
  memcpy(&memid,seg,4);

  printf("%d memid\n",memid);

  if(memid != -1){
    // IPC_PRIVATE Memory Segement
    if(memid == 0x0){
      printf("No memory segment identifier set\nAllocating Private Segment");
      shmid = shmget(IPC_PRIVATE, memsize, SHM_MODE);
      if(shmid>-1){
	shmptr = (char*) shmat(shmid, 0,0);
      }else{
	printf("Failed to attach or create a PRIVATE shared memory segment");
	exit(FAIL);
      } // endif shmid
    } // endif IPC_PRIVATE


    // Public Named Segment    
    if(memid != 0x0){
      // First Check to see if the segment already exists and if it does get 
      // the stats on it so we can check it's current size
      struct shmid_ds shm_seg_info;

      printf("Attempting to connect to shared memory\nID: 0x%08x\n",memid);

      if(shmid >= 0){
	if(shmid == 0){
	  // Sigh we should never have a zero ID for the segment, but on the old TDU linux we 
	  // do get a zero id the very first time we try to make a segment after a reboot
	  printf("Received shared memory segment with ID = 0.  This is not standard, but tolerated\n");
	}
	// Check that we got an existing segment
	// Retrieve the information on the segment (mainly the size)
	int result = shmctl(shmid, IPC_STAT, &shm_seg_info);
	if(result < 0){
	  perror("Unable to stat shared memory segment"); 
	  exit(-1);
	}
	// Check the existing size of the segment and compare it against
	// the requested size
	if( shm_seg_info.shm_segsz < memsize){
	  // The existing segment has a size that is less than the requested size
	  printf("Existing Shared Memory segment has a size less than the requested minimum (Requested: %u, Actual: %zu\n)\n",
		 memsize,shm_seg_info.shm_segsz);
	};
	shmptr = (char*) shmat(shmid, 0,0);
      }else{
	printf("Failed to attach to existing shared memory segment using:\nID: %d\tSize:%d\n",memid,memsize);
	printf("Creating new segment with Identifier: 0x%08x, Size: %u bytes\n",memid,memsize);
	shmid = shmget(memid, memsize, SHM_MODE_CREAT);
	if(shmid){
	  shmptr = (char*) shmat(shmid, 0,0);
	}else{
	  printf("Failed to attach or create a shared memory segmetn using:\nID: %d\tSize:%d\n",memid,memsize);
	  exit(FAIL);
	} // endif shmid
      } // endif shmid > 0      
            
      shmptr_start   = shmptr;
      shmptr_int     = (int*) shmptr;
      shmptr_end     = shmptr_start + memsize;
      shmptr_current = shmptr_start;
    } // endif memid != 0x0
  } // endif memid != -1


  DataMap   = (SHMSpillDataMap*)shmptr_start;
  AccelData = (AccelEventData*) (DataMap + 2);

  (*DataMap).data_start   = (uint32_t*) AccelData;
  (*DataMap).data_end     = (uint32_t*) shmptr_end;
  (*DataMap).data_insert  = AccelData;
  (*DataMap).data_entries = 0;


  /************************************************************/

  std::cout << "Sending spills..." << std::endl;

  while (true) {

    int bytesRead = mtu.readEvents((uint8_t *)buffer, 4096);

    //printf ("bytes read = %d\n", bytesRead);
    if (bytesRead%10 != 0) {
      std::cout << "I expected to read an integer multiple of 10 bytes.\n"
                << "I got " << bytesRead << " bytes instead.\n"
                << "This is weird. Alert an expert!!"<<std::endl;
    }

    uint64_t latestTCR = 0x0; // Keep track of the TCR events
    uint64_t oneSecondBoundary = 0;
    uint64_t oneSecondBoundary2 = 0;
    int64_t deltaT  = 0;
    int64_t deltaT1 = 0;
    int64_t deltaT2 = 0;
    const uint64_t ONE_SECOND = 64000000;


    for (int i=0; i<bytesRead/10; ++i)
    {
      uint16_t evt = buffer[i*5+0];
      uint64_t novatime = 0;
      for (int j=0; j!=4; ++j) {
        novatime += ((uint64_t)buffer[i*5+(j+1)])<<(16*j);
      }

      novatime = novatime >> 1;

      //Subtract 15 seconds if we run with old firmware
      if (subfifteen) novatime -= fifteensecs;

      //decode time into spillInfoXml
      NssUtil::gpstimeToString(novatime, str);
      spillInfoXml[0] = str;

      //decode evt into spillInfoXml
      NssSpillInfo::getSpillTypeFromEvent(spillType,evt);
      spillInfoXml[1] = spillType;
      
      // Check for the TCR Reference Signals
      if (evt == 0x041B){
	// We don't issue triggers on the references, instead we check thier offset 
	// from the canonical 1 second GPS boundary

	latestTCR = novatime;

	oneSecondBoundary  = novatime - (novatime % ONE_SECOND);   // Compute the 1 second boundary
        oneSecondBoundary2 = novatime - (novatime % ONE_SECOND) + ONE_SECOND; //nearest one second boundary could be in the future, check for that
	deltaT1            = llabs(latestTCR - oneSecondBoundary);        // Compute the time delta
        deltaT2            = llabs(latestTCR - oneSecondBoundary2); //Compute the delta for the next second

	// Report the smaller time difference
        deltaT = ((deltaT1 < deltaT2) ? deltaT1 : deltaT2); 

	// Check if the time delta is within some threshold of the 1 second boundary
	if(llabs(deltaT) > (int64_t)tappCfg.TCRErrorThreshold()){
	  // If it is not then issue errors to the system
	  ::mf::LogError("TimePhaseDifference") << "TDU Drift ERROR: " << hostname << "-->" 
						<< " Timestamp Delta from Reference: " << deltaT 
						<< " at Novatime: " << novatime 
						<< std::endl;  
	}else{
	  ::mf::LogInfo("TimePhaseDifference") << "TDU drift Nominal: " << hostname << "-->" 
					       << " Timestamp Delta from Reference: " << deltaT 
					       << " at Novatime: " << novatime 
					       << std::endl;  
	  
	}

        // wait 300 TCR triggers (5 minutes) before checking for jumps in the trigger.  
	// This is to get over the initial backlog of triggers in the buffer when the server starts
        if (tcrCount < 300) {
          tcrCount++;
        }
        else {
          deltaT = latestTCR - lastTCR_time - ONE_SECOND; //compare the difference between the current and previous TCR trigger to a 1 second interval
	  if(llabs(deltaT) > (int64_t)tappCfg.TCRErrorThreshold()){
	    // If it is not then issue errors to the system
	    ::mf::LogError("TimeReferenceInterval") << "TDU Ref Interval ERROR: " << hostname << "-->" 
						    << "TCR reference interval: " << (int64_t)latestTCR - (int64_t)lastTCR_time
						    << "( current: " << latestTCR
						    << " previous: " << lastTCR_time << " )"
						    << std::endl;  
	  }else{
	    ::mf::LogInfo("TimeReferenceInterval") << "TDU Sync Nominal: " << hostname << "-->"
						    << "TCR reference interval: " << (int64_t)latestTCR - (int64_t)lastTCR_time
						    << "( current: " << latestTCR
						    << " previous: " << lastTCR_time << " )"
						    << std::endl;  	  
	  }
        }


        lastTCR_time = novatime;



	// Record the trigger information to our shared memory segment

	// First copy the information into a structure
	theAccelData.evt_type = spillType;
	theAccelData.time     = novatime;
	theAccelData.evt_no++;

	// Now to a single copy into the shared memory area
	memcpy((*DataMap).data_insert,&theAccelData,sizeof(theAccelData));

#if 0
	printf("SHM Start: 0x%08x\tDM Insert: 0x%08x\tAccelDat: 0x%08x\tSize: %u\n",
	       shmptr_start,
	       (*DataMap).data_insert,
	       &theAccelData,
	       sizeof(theAccelData));
#endif

	// Update the insertion points to point to the next area or back to the
	// start of the data buffer
	(*DataMap).data_insert++;
  	if((uint32_t*)(*DataMap).data_insert > (*DataMap).data_end){
  	  (*DataMap).data_insert = (struct AccelEventData*)(*DataMap).data_start;
	}

  	(*DataMap).data_entries++;

      }
    }
    sleep(1);
  }
 
  return 0;
}
